from PySide6.QtCore import QSettings
from airunner.settings import AIRUNNER_ORGANIZATION, AIRUNNER_APPLICATION_NAME


def get_qsettings() -> QSettings:
    return QSettings(AIRUNNER_ORGANIZATION, AIRUNNER_APPLICATION_NAME)



from PySide6.QtCore import Slot
from airunner.gui.windows.setup_wizard.base_wizard import BaseWizard


class DownloadWizardPage(BaseWizard):
    def __init__(self, *args):
        super(DownloadWizardPage, self).__init__(*args)
        self.download_thread = None
        self.models_to_download = []
        if self.ui:
            self.ui.download.clicked.connect(self.download_models)
        self._original_text = self.ui.download.text()
        self.downloading = False
        self.ui.download.show()
        self.ui.cancel.hide()

    @Slot()
    def download_models(self):
        if not self.downloading:
            self.downloading = True
            if self.ui:
                self.ui.download.setText("Cancel")
                self.ui.download.hide()
                self.ui.cancel.show()
            self.start_download()

    def start_download(self):
        """
        Placeholder for download function
        :return:
        """
        pass

    @Slot()
    def cancel(self):
        self.downloading = False
        if self.ui:
            self.ui.download.setText(self._original_text)
            self.ui.download.show()
            self.ui.cancel.hide()

    def update_progress(self, current, total):
        for data in self.models_to_download:
            if "progress_bar" in data:
                progress_bar = data["progress_bar"]
                progress_bar.setMaximum(total)
                progress_bar.setValue(current)

    def download_finished(self):
        for data in self.models_to_download:
            if "tqdm_class" in data:
                data["tqdm_class"].clear()

    def closeEvent(self, event):
        if self.download_thread.isRunning():
            self.download_thread.stop()
            self.download_thread.wait()
        event.accept()



from tqdm import tqdm


class CustomTqdmProgressBar:
    def __init__(self, progress_bar):
        self.progress_bar = progress_bar
        self._lock = None

    def __call__(
        self,
        iterable=None,
        desc=None,
        total=None,
        leave=True,
        file=None,
        ncols=None,
        mininterval=0.1,
        maxinterval=10.0,
        miniters=None,
        ascii=None,
        disable=False,
        unit='it',
        unit_scale=False,
        dynamic_ncols=False,
        smoothing=0.3,
        bar_format=None,
        initial=0,
        position=None,
        postfix=None,
        unit_divisor=1000,
        write_bytes=None,
        lock_args=None,
        nrows=None,
        colour=None,
        delay=0,
        gui=False,
        **kwargs
    ):
        if total is None:
            total = self.progress_bar.maximum()
        self.tqdm_instance = tqdm(
            iterable=iterable,
            desc=desc,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            disable=disable,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs
        )
        return self.tqdm_instance

    @property
    def n(self):
        return self.tqdm_instance.n

    @property
    def total(self):
        return self.tqdm_instance.total

    def update(self, n=1):
        self.tqdm_instance.update(n)
        self.progress_bar.setValue(self.tqdm_instance.n)

    def close(self):
        self.tqdm_instance.close()
        self.progress_bar.setValue(self.progress_bar.maximum())

    def set_lock(self, lock):
        self._lock = lock

    def get_lock(self):
        return self._lock

    def clear(self):
        if hasattr(self, '_lock'):
            del self._lock







# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'lora.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGridLayout, QLineEdit,
    QPushButton, QSizePolicy, QWidget)

from airunner.gui.widgets.slider.slider_widget import SliderWidget
import airunner.feather_rc

class Ui_lora(object):
    def setupUi(self, lora):
        if not lora.objectName():
            lora.setObjectName(u"lora")
        lora.resize(518, 80)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(lora.sizePolicy().hasHeightForWidth())
        lora.setSizePolicy(sizePolicy)
        lora.setMaximumSize(QSize(16777215, 16777215))
        font = QFont()
        font.setPointSize(9)
        lora.setFont(font)
        self.gridLayout_2 = QGridLayout(lora)
        self.gridLayout_2.setSpacing(0)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setHorizontalSpacing(0)
        self.lora_container = QWidget(lora)
        self.lora_container.setObjectName(u"lora_container")
        self.gridLayout_3 = QGridLayout(self.lora_container)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setHorizontalSpacing(5)
        self.gridLayout_3.setVerticalSpacing(10)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.enabledCheckbox = QCheckBox(self.lora_container)
        self.enabledCheckbox.setObjectName(u"enabledCheckbox")

        self.gridLayout_3.addWidget(self.enabledCheckbox, 0, 0, 1, 1)

        self.trigger_word_edit = QLineEdit(self.lora_container)
        self.trigger_word_edit.setObjectName(u"trigger_word_edit")

        self.gridLayout_3.addWidget(self.trigger_word_edit, 2, 0, 1, 1)

        self.delete_button = QPushButton(self.lora_container)
        self.delete_button.setObjectName(u"delete_button")
        self.delete_button.setMinimumSize(QSize(24, 24))
        self.delete_button.setMaximumSize(QSize(24, 24))
        self.delete_button.setCursor(QCursor(Qt.PointingHandCursor))
        icon = QIcon()
        icon.addFile(u":/dark/icons/feather/dark/trash-2.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.delete_button.setIcon(icon)

        self.gridLayout_3.addWidget(self.delete_button, 2, 2, 1, 1)

        self.scale_slider = SliderWidget(self.lora_container)
        self.scale_slider.setObjectName(u"scale_slider")
        sizePolicy.setHeightForWidth(self.scale_slider.sizePolicy().hasHeightForWidth())
        self.scale_slider.setSizePolicy(sizePolicy)
        self.scale_slider.setMinimumSize(QSize(0, 0))
        self.scale_slider.setProperty("slider_minimum", 0)
        self.scale_slider.setProperty("slider_maximum", 100)
        self.scale_slider.setProperty("spinbox_minimum", 0.000000000000000)
        self.scale_slider.setProperty("spinbox_maximum", 1.000000000000000)
        self.scale_slider.setProperty("display_as_float", True)
        self.scale_slider.setProperty("slider_single_step", 1)
        self.scale_slider.setProperty("slider_page_step", 5)
        self.scale_slider.setProperty("spinbox_single_step", 0.010000000000000)
        self.scale_slider.setProperty("spinbox_page_step", 0.100000000000000)
        self.scale_slider.setProperty("table_id", 0)

        self.gridLayout_3.addWidget(self.scale_slider, 1, 0, 1, 3)


        self.gridLayout.addWidget(self.lora_container, 0, 0, 2, 1)


        self.gridLayout_2.addLayout(self.gridLayout, 0, 0, 1, 1)


        self.retranslateUi(lora)
        self.trigger_word_edit.textChanged.connect(lora.action_changed_trigger_words)
        self.enabledCheckbox.toggled.connect(lora.action_toggled_lora_enabled)
        self.delete_button.clicked.connect(lora.action_clicked_button_deleted)

        QMetaObject.connectSlotsByName(lora)
    # setupUi

    def retranslateUi(self, lora):
        lora.setWindowTitle(QCoreApplication.translate("lora", u"Form", None))
        self.enabledCheckbox.setText(QCoreApplication.translate("lora", u"enabledCheckbox", None))
#if QT_CONFIG(tooltip)
        self.trigger_word_edit.setToolTip(QCoreApplication.translate("lora", u"<html><head/><body><p>Some LoRA require a trigger word to activate.</p><p>Make a note here for your records.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.trigger_word_edit.setPlaceholderText(QCoreApplication.translate("lora", u"Trigger words (comma separated)", None))
#if QT_CONFIG(tooltip)
        self.delete_button.setToolTip(QCoreApplication.translate("lora", u"Delete model", None))
#endif // QT_CONFIG(tooltip)
        self.delete_button.setText("")
        self.scale_slider.setProperty("settings_property", QCoreApplication.translate("lora", u"lora.scale", None))
        self.scale_slider.setProperty("label_text", QCoreApplication.translate("lora", u"Scale", None))
    # retranslateUi





import torch
from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication
from airunner.enums import SignalCode, ModelStatus, ModelType, StatusColors
from airunner.gui.widgets.base_widget import BaseWidget
from airunner.gui.widgets.status.templates.status_ui import Ui_status_widget
from airunner.settings import AIRUNNER_ART_ENABLED


class StatusWidget(BaseWidget):
    widget_class_ = Ui_status_widget

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.APPLICATION_STATUS_INFO_SIGNAL: self.on_status_info_signal,
            SignalCode.APPLICATION_STATUS_ERROR_SIGNAL: self.on_status_error_signal,
            SignalCode.APPLICATION_CLEAR_STATUS_MESSAGE_SIGNAL: self.on_clear_status_message_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.SD_PIPELINE_LOADED_SIGNAL: self.set_sd_pipeline_label,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed,
        }
        super().__init__(*args, **kwargs)

        if not AIRUNNER_ART_ENABLED:
            self.ui.sd_status.deleteLater()
            self.ui.controlnet_status.deleteLater()
            self.ui.nsfw_status.deleteLater()

        self.set_sd_pipeline_label({"pipeline": ""})
        self.version = None

        self.timer = QTimer()
        self.timer.timeout.connect(self.update_system_stats)
        self.timer.start(1000)

        self.safety_checker_status = ModelStatus.UNLOADED
        self.feature_extractor_status = ModelStatus.UNLOADED
        self._model_status = {model_type: ModelStatus.UNLOADED for model_type in ModelType}

        if self.application_settings.nsfw_filter and self.application_settings.sd_enabled:
            self.safety_checker_status = ModelStatus.LOADING
            self.feature_extractor_status = ModelStatus.LOADING

        self.update_system_stats()

        for item in [
            (ModelType.LLM, "llm_enabled"),
            (ModelType.TTS, "tts_enabled"),
            (ModelType.STT, "stt_enabled"),
            (ModelType.SD, "sd_enabled"),
            (ModelType.CONTROLNET, "controlnet_enabled"),
        ]:
            if getattr(self.application_settings, item[1]):
                self._model_status[item[0]] = ModelStatus.LOADING
                self.update_model_status({
                    "model": item[0],
                    "status": ModelStatus.LOADING
                })
    
    def on_application_settings_changed(self):
        self.set_sd_status_text()
    
    def set_sd_pipeline_label(self, data):
        if data["pipeline"]:
            self.ui.pipeline_label.setText(data["pipeline"])
            self.ui.pipeline_divider.show()
        else:
            self.ui.pipeline_label.setText("")
            self.ui.pipeline_divider.hide()

    def showEvent(self, event):
        super().showEvent(event)
        self.set_sd_status_text()

        if self.application_settings.sd_enabled:
            self.on_model_status_changed_signal({
                "model": ModelType.SD,
                "status": self._model_status[ModelType.SD],
                "path": ""
            })
        if self.application_settings.controlnet_enabled:
            self.on_model_status_changed_signal({
                "model": ModelType.CONTROLNET,
                "status": self._model_status[ModelType.CONTROLNET],
                "path": ""
            })
        if self.application_settings.llm_enabled:
            self.on_model_status_changed_signal({
                "model": ModelType.LLM,
                "status": self._model_status[ModelType.LLM],
                "path": ""
            })
        if self.application_settings.tts_enabled:
            self.on_model_status_changed_signal({
                "model": ModelType.TTS,
                "status": self._model_status[ModelType.TTS],
                "path": ""
            })
        if self.application_settings.stt_enabled:
            self.on_model_status_changed_signal({
                "model": ModelType.STT,
                "status": self._model_status[ModelType.STT],
                "path": ""
            })
        if self.application_settings.nsfw_filter and self.application_settings.sd_enabled:
            self.on_model_status_changed_signal({
                "model": ModelType.SAFETY_CHECKER,
                "status": self._model_status[ModelType.SAFETY_CHECKER],
                "path": ""
            })

    def update_model_status(self, data):
        self._model_status[data["model"]] = data["status"]
        if data["status"] is ModelStatus.LOADING:
            color = StatusColors.LOADING
        elif data["status"] is ModelStatus.LOADED:
            color = StatusColors.LOADED
        elif data["status"] is ModelStatus.FAILED:
            color = StatusColors.FAILED
        elif data["status"] is ModelStatus.READY:
            color = StatusColors.READY
        else:
            color = StatusColors.UNLOADED

        styles = "QLabel { color: " + color.value + "; }"
        element_name = ""
        tool_tip = ""
        if not data["model"]:
            return
        if data["model"] == ModelType.SD:
            element_name = "sd_status"
            tool_tip = "Stable Diffusion"
            self.set_sd_status_text()
        elif data["model"] == ModelType.CONTROLNET:
            element_name = "controlnet_status"
            tool_tip = "Controlnet"
        elif data["model"] == ModelType.LLM:
            element_name = "llm_status"
            tool_tip = "LLM"
        elif data["model"] == ModelType.TTS:
            element_name = "tts_status"
            tool_tip = "TTS"
        elif data["model"] == ModelType.STT:
            element_name = "stt_status"
            tool_tip = "STT"
        elif data["model"] == ModelType.SAFETY_CHECKER:
            element_name = "nsfw_status"
            tool_tip = "Safety Checker"

        tool_tip += " " + data["status"].value

        if element_name != "":
            getattr(self.ui, element_name).setStyleSheet(styles)
            getattr(self.ui, element_name).setToolTip(tool_tip)
        QApplication.processEvents()

    def set_sd_status_text(self):
        if self.version != self.generator_settings.version:
            version = self.generator_settings.version
            try:
                self.ui.sd_status.setText(version)
            except RuntimeError as e:
                if AIRUNNER_ART_ENABLED:
                    self.logger.warning(f"Error setting SD status text: {e}")

    def on_model_status_changed_signal(self, data):
        self.update_model_status(data)

    def on_status_info_signal(self, message):
        self.set_system_status(message, error=False)

    def on_status_error_signal(self, message):
        self.set_system_status(message, error=True)

    def on_clear_status_message_signal(self):
        self.set_system_status("", error=False)

    def update_system_stats(self):
        cuda_status = f"{'NVIDIA' if torch.cuda.is_available() else 'CPU'}"

        # Color by has_cuda red for disabled, green for enabled
        color = StatusColors.LOADED if torch.cuda.is_available() else StatusColors.FAILED
        self.ui.cuda_status.setStyleSheet(
            "QLabel { color: " + color.value + "; }"
        )

        self.ui.cuda_status.setText(cuda_status)

    def set_system_status(self, txt, error):
        if type(txt) is dict:
            txt = ""
        self.ui.system_message.setText(txt)
        if error:
            self.ui.system_message.setStyleSheet("QLabel { color: #ff0000; }")
        else:
            self.ui.system_message.setStyleSheet("QLabel { color: #ffffff; }")
        QApplication.processEvents()



from sqlalchemy import Column, Integer, String, Boolean
from PySide6.QtWidgets import QTabWidget

from airunner.data.models.base import BaseModel
from airunner.data.session_manager import session_scope


class Tab(BaseModel):
    __tablename__ = 'tabs'
    id = Column(Integer, primary_key=True, autoincrement=True)
    section = Column(String, default="", nullable=False)
    name = Column(String, default="", nullable=False)
    active = Column(Boolean, default=False, nullable=False)
    displayed = Column(Boolean, default=True, nullable=False)
    index = Column(Integer, default=0, nullable=False)

    @staticmethod
    def update_tabs(section, tab_widget, index):
        """Update tabs in the database based on the active tab index."""
        # Ensure the index is valid before starting a session
        if index < 0 or index >= tab_widget.count():
            raise IndexError("Tab index out of range.")

        with session_scope() as session:
            # Set all tabs in the section to inactive
            session.query(Tab).filter(Tab.section == section).update({"active": False})

            # Get the tab name at the given index
            tab_name = tab_widget.tabText(index)

            # Set the selected tab to active
            session.query(Tab).filter(Tab.section == section, Tab.name == tab_name).update({"active": True})
            session.commit()



from PySide6.QtWidgets import QWidget
from abc import ABC, abstractmethod


class BasePlugin(ABC):
    name: str = ""

    @abstractmethod
    def get_widget(self) -> QWidget:
        """
        Get the widget for the plugin.

        Returns:
            QWidget: The widget for the plugin.
        """


import io

import PIL
from PIL import Image


def convert_binary_to_image(binary_data: bytes) -> Image:
    if binary_data is None:
        return None
    try:
        bytes_ = io.BytesIO(binary_data)
        return Image.open(bytes_)
    except PIL.UnidentifiedImageError as e:
        print(f"Something went wrong with binary data conversion to image: {e}")
        return None
    except TypeError as e:
        print(f"Something went wrong with binary data conversion to image: {e}")
        return None



import subprocess


def bash_execute(command: str) -> str:
    """
    Executes a bash command.

    This is an extremely unsafe operation and should be used with caution.
    This allows an LLM to execute arbitrary bash commands.
    By default this is disabled in the settings.

    :param command: str
    :return: str
    """
    try:
        command = command.split(" ")
        result = subprocess.check_output(command, shell=False)
        return result.decode("utf-8")
    except Exception as e:
        return str(e)



"""Tests for the image filters package.

This module contains unit tests for the filters in the airunner.filters package.
"""

import unittest
from unittest.mock import MagicMock, patch
import numpy as np
from PIL import Image

from airunner.filters import (
    BaseFilter,
    BoxBlur,
    ColorBalanceFilter,
    Dither,
    FilmFilter,
    GaussianBlur,
    HalftoneFilter,
    Invert,
    PixelFilter,
    RegistrationErrorFilter,
    RGBNoiseFilter,
    SaturationFilter,
    UnsharpMask
)


class TestBaseFilter(unittest.TestCase):
    """Test the BaseFilter class functionality."""
    
    def setUp(self):
        """Set up test fixtures."""
        self.test_image = Image.new('RGB', (100, 100), color='white')
    
    def test_init(self):
        """Test filter initialization with kwargs."""
        test_filter = BaseFilter(param1='value1', param2='value2')
        self.assertEqual(test_filter.param1, 'value1')
        self.assertEqual(test_filter.param2, 'value2')
    
    def test_filter_caching(self):
        """Test image caching behavior."""
        # Create a test subclass
        class TestFilter(BaseFilter):
            def apply_filter(self, image, do_reset=False):
                self.apply_filter_called = True
                self.do_reset_passed = do_reset
                return image
        
        test_filter = TestFilter()
        test_filter.apply_filter_called = False
        test_filter.do_reset_passed = None
        
        # First call should set do_reset to True
        test_filter.filter(self.test_image)
        self.assertTrue(test_filter.apply_filter_called)
        self.assertTrue(test_filter.do_reset_passed)
        
        # Reset for next test
        test_filter.apply_filter_called = False
        test_filter.do_reset_passed = None
        
        # Second call with same image should set do_reset to False
        test_filter.filter(self.test_image)
        self.assertTrue(test_filter.apply_filter_called)
        self.assertFalse(test_filter.do_reset_passed)
        
        # Reset for next test
        test_filter.apply_filter_called = False
        test_filter.do_reset_passed = None
        
        # Different image should set do_reset to True again
        new_image = Image.new('RGB', (100, 100), color='black')
        test_filter.filter(new_image)
        self.assertTrue(test_filter.apply_filter_called)
        self.assertTrue(test_filter.do_reset_passed)


class TestBoxBlur(unittest.TestCase):
    """Test the BoxBlur filter."""
    
    def setUp(self):
        """Set up test fixtures."""
        self.test_image = Image.new('RGB', (100, 100), color='white')
    
    def test_apply_filter(self):
        """Test the box blur filter application."""
        filter_instance = BoxBlur(radius=2)
        result = filter_instance.filter(self.test_image)
        
        # The result should be an image
        self.assertIsInstance(result, Image.Image)
        
        # The dimensions should remain the same
        self.assertEqual(result.size, self.test_image.size)


class TestInvert(unittest.TestCase):
    """Test the Invert filter."""
    
    def setUp(self):
        """Set up test fixtures."""
        self.test_image = Image.new('RGB', (100, 100), color='white')
    
    def test_apply_filter(self):
        """Test the invert filter application."""
        filter_instance = Invert()
        result = filter_instance.filter(self.test_image)
        
        # The result should be an image
        self.assertIsInstance(result, Image.Image)
        
        # For a white image, inversion should result in a black image
        # Get the pixel at (0,0)
        pixel = result.getpixel((0, 0))
        # All RGB values should be 0 (black)
        self.assertEqual(pixel[0], 0)
        self.assertEqual(pixel[1], 0)
        self.assertEqual(pixel[2], 0)


class TestRGBNoiseFilter(unittest.TestCase):
    """Test the RGBNoiseFilter."""
    
    def setUp(self):
        """Set up test fixtures."""
        self.test_image = Image.new('RGBA', (100, 100), color='white')
    
    @patch('numpy.random.uniform')
    def test_apply_filter(self, mock_uniform):
        """Test the RGB noise filter application."""
        # Mock the random function to return a predictable value
        mock_uniform.return_value = np.ones((100, 100))
        
        # Create filter with specific noise values
        filter_instance = RGBNoiseFilter(red=10, green=20, blue=30)
        result = filter_instance.filter(self.test_image)
        
        # The result should be an image
        self.assertIsInstance(result, Image.Image)
        
        # Check that numpy.random.uniform was called correctly
        mock_uniform.assert_called()


class TestDither(unittest.TestCase):
    """Test the Dither filter."""
    
    def setUp(self):
        """Set up test fixtures."""
        self.test_image = Image.new('RGB', (100, 100), color='white')
    
    def test_apply_filter(self):
        """Test the dither filter application."""
        filter_instance = Dither(threshold=0.5)
        result = filter_instance.filter(self.test_image)
        
        # The result should be an image
        self.assertIsInstance(result, Image.Image)
        
        # Mode should be '1' (binary image)
        self.assertEqual(result.mode, '1')


if __name__ == '__main__':
    unittest.main()


from typing import Dict
import threading

import torch
from PySide6.QtCore import QThread
from PySide6.QtCore import QObject, Signal, Slot

from airunner.enums import QueueType, SignalCode, ModelType, ModelAction
from airunner.workers.worker import Worker
from airunner.handlers import StableDiffusionModelManager
from airunner.handlers.stablediffusion.image_request import ImageRequest
from airunner.settings import (
    AIRUNNER_SD_ON,
    AIRUNNER_ART_MODEL_PATH,
    AIRUNNER_ART_MODEL_VERSION,
    AIRUNNER_ART_PIPELINE,
    AIRUNNER_ART_SCHEDULER,
    AIRUNNER_ART_USE_COMPEL,
)

torch.backends.cuda.matmul.allow_tf32 = True


class GenerateWorker(QObject):
    finished = Signal()
    error = Signal(str)

    def __init__(self, sd, message):
        super().__init__()
        self.sd = sd
        self.message = message

    @Slot()
    def run(self):
        try:
            self.sd.handle_generate_signal(self.message)
        except Exception as e:
            self.error.emit(str(e))
        finally:
            self.finished.emit()


class SDWorker(Worker):
    queue_type = QueueType.GET_LAST_ITEM

    def __init__(self):
        self.sd = None
        self.signal_handlers = {
            SignalCode.SD_CANCEL_SIGNAL: self.on_sd_cancel_signal,
            SignalCode.START_AUTO_IMAGE_GENERATION_SIGNAL: self.on_start_auto_image_generation_signal,
            SignalCode.STOP_AUTO_IMAGE_GENERATION_SIGNAL: self.on_stop_auto_image_generation_signal,
            SignalCode.DO_GENERATE_SIGNAL: self.on_do_generate_signal,
            SignalCode.INTERRUPT_IMAGE_GENERATION_SIGNAL: self.on_interrupt_image_generation_signal,
            SignalCode.CHANGE_SCHEDULER_SIGNAL: self.on_change_scheduler_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.SD_LOAD_SIGNAL: self.on_load_stablediffusion_signal,
            SignalCode.SD_UNLOAD_SIGNAL: self.on_unload_stablediffusion_signal,
            SignalCode.CONTROLNET_LOAD_SIGNAL: self.on_load_controlnet_signal,
            SignalCode.CONTROLNET_UNLOAD_SIGNAL: self.on_unload_controlnet_signal,
            SignalCode.LORA_UPDATE_SIGNAL: self.on_update_lora_signal,
            SignalCode.EMBEDDING_UPDATE_SIGNAL: self.on_update_embeddings_signal,
            SignalCode.EMBEDDING_DELETE_MISSING_SIGNAL: self.delete_missing_embeddings,
            SignalCode.SAFETY_CHECKER_LOAD_SIGNAL: self.on_load_safety_checker,
            SignalCode.SAFETY_CHECKER_UNLOAD_SIGNAL: self.on_unload_safety_checker,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed,
        }
        super().__init__()
        self.__requested_action = ModelAction.NONE
        self._threads = []
        self._workers = []

    def on_load_safety_checker(self):
        if self.sd:
            thread = threading.Thread(target=self._load_safety_checker)
            thread.start()

    def on_unload_safety_checker(self):
        if self.sd:
            thread = threading.Thread(target=self._unload_safety_checker)
            thread.start()
    
    def on_application_settings_changed(self):
        if self.sd:
            self.sd.on_application_settings_changed()

    def scan_for_embeddings(self):
        if self.sd:
            self.sd.scan_for_embeddings()

    def delete_missing_embeddings(self, message):
        if self.sd:
            self.sd.delete_missing_embeddings(message)

    def get_embeddings(self, message):
        if self.sd:
            self.sd.get_embeddings(message)

    def on_update_lora_signal(self):
        thread = threading.Thread(target=self._reload_lora)
        thread.start()

    def _reload_lora(self):
        if self.sd:
            self.sd.reload_lora()

    def on_update_embeddings_signal(self):
        if self.sd:
            self.sd.reload_embeddings()

    def on_add_lora_signal(self, message):
        if self.sd:
            self.sd.on_add_lora_signal(message)

    def on_load_controlnet_signal(self, _data=None):
        if self.sd:
            thread = threading.Thread(target=self._load_controlnet)
            thread.start()

    def on_unload_controlnet_signal(self, _data=None):
        if self.sd:
            thread = threading.Thread(target=self._unload_controlnet)
            thread.start()

    def on_load_stablediffusion_signal(self, data: Dict = None):
        if self.sd:
            thread = threading.Thread(target=self._load_sd, args=(data,))
            thread.start()

    def on_unload_stablediffusion_signal(self, data=None):
        if self.sd:
            thread = threading.Thread(target=self._unload_sd, args=(data,))
            thread.start()

    def _load_sd(self, data: Dict = None):
        do_reload = data.get("do_reload", False)
        if do_reload:
            self.sd.reload()
        else:
            self.sd.load()
        if data:
            callback = data.get("callback", None)
            if callback is not None:
                callback(data)

    def _unload_sd(self, data: Dict = None):
        self.sd.unload()
        if data:
            callback = data.get("callback", None)
            if callback is not None:
                callback(data)

    def _load_controlnet(self):
        self.sd.load_controlnet()

    def _unload_controlnet(self):
        self.sd.unload_controlnet()

    def _load_safety_checker(self):
        self.sd.load_safety_checker()

    def _unload_safety_checker(self):
        self.sd.unload_safety_checker()

    def on_tokenizer_load_signal(self, data: Dict = None):
        if self.sd:
            self.sd.sd_load_tokenizer(data)

    def start_worker_thread(self):
        generator_settings = self.generator_settings
        model_path = generator_settings.aimodel.path if generator_settings.aimodel else ""
        if AIRUNNER_ART_MODEL_PATH != "":
            model_path = AIRUNNER_ART_MODEL_PATH
        
        model_version = generator_settings.version if generator_settings.version else ""
        if AIRUNNER_ART_MODEL_VERSION != "":
            model_version = AIRUNNER_ART_MODEL_VERSION
        
        pipeline = generator_settings.pipeline_action if generator_settings.pipeline_action else ""
        if AIRUNNER_ART_PIPELINE != "":
            pipeline = AIRUNNER_ART_PIPELINE
        
        scheduler_name = generator_settings.scheduler if generator_settings.scheduler else ""
        if AIRUNNER_ART_SCHEDULER != "":
            scheduler_name = AIRUNNER_ART_SCHEDULER
        
        use_compel = generator_settings.use_compel
        if AIRUNNER_ART_USE_COMPEL != "":
            use_compel = AIRUNNER_ART_USE_COMPEL

        self.sd = StableDiffusionModelManager(
            model_path=model_path,
            model_version=model_version,
            pipeline=pipeline,
            scheduler_name=scheduler_name,
            use_compel=use_compel
        )
        
        if self.application_settings.sd_enabled or AIRUNNER_SD_ON:
            self.sd.load()

    def handle_message(self, message):
        if self.sd:
            self.sd.run()

    @staticmethod
    def on_sd_cancel_signal(_data=None):
        print("on_sd_cancel_signal")

    def on_start_auto_image_generation_signal(self, _data=None):
        pass

    def on_stop_auto_image_generation_signal(self, _data=None):
        # self.sd_mode = SDMode.STANDARD
        pass

    def on_do_generate_signal(self, message: Dict):
        if self.sd:
            if not message.get("image_request", None):
                settings = self.generator_settings
                message["image_request"] = ImageRequest(
                    pipeline_action=settings.pipeline_action,
                    generator_name=settings.generator_name,
                    prompt=settings.prompt,
                    negative_prompt=settings.negative_prompt,
                    second_prompt=settings.second_prompt,
                    second_negative_prompt=settings.second_negative_prompt,
                    random_seed=settings.random_seed,
                    model_path=settings.aimodel.path if settings.aimodel else "",
                    scheduler=settings.scheduler,
                    version=settings.version,
                    use_compel=settings.use_compel,
                    steps=settings.steps,
                    ddim_eta=settings.ddim_eta,
                    scale=settings.scale / 100,
                    seed=settings.seed,
                    strength=settings.strength / 100,
                    n_samples=settings.n_samples,
                    clip_skip=settings.clip_skip,
                    crops_coord_top_left=settings.crops_coord_top_left,
                    original_size=settings.original_size,
                    target_size=settings.target_size,
                    negative_original_size=settings.negative_original_size,
                    negative_target_size=settings.negative_target_size,
                    lora_scale=settings.lora_scale,
                )
            thread = QThread()
            worker = GenerateWorker(self.sd, message)
            worker.moveToThread(thread)
            thread.started.connect(worker.run)
            worker.finished.connect(thread.quit)
            worker.finished.connect(worker.deleteLater)
            thread.finished.connect(thread.deleteLater)
            worker.error.connect(self.handle_error)
            self._threads.append(thread)
            self._workers.append(worker)
            thread.start()

    @staticmethod
    def handle_error(error_message):
        import traceback
        traceback.print_stack()
        print(f"SDWorker Error: {error_message}")

    def on_interrupt_image_generation_signal(self, _data=None):
        if self.sd:
            self.sd.interrupt_image_generation()

    def on_change_scheduler_signal(self, data: Dict):
        if self.sd:
            self.sd.load_scheduler(data["scheduler"])

    def on_model_status_changed_signal(self, message: Dict):
        if self.sd and message["model"] == ModelType.SD:
            if self.__requested_action is ModelAction.CLEAR:
                self.on_unload_stablediffusion_signal()
            self.__requested_action = ModelAction.NONE



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'stable_diffusion_model_setup.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFrame, QGridLayout,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QWidget)

class Ui_stable_diffusion_model_setup(object):
    def setupUi(self, stable_diffusion_model_setup):
        if not stable_diffusion_model_setup.objectName():
            stable_diffusion_model_setup.setObjectName(u"stable_diffusion_model_setup")
        stable_diffusion_model_setup.resize(447, 533)
        self.gridLayout = QGridLayout(stable_diffusion_model_setup)
        self.gridLayout.setObjectName(u"gridLayout")
        self.frame = QFrame(stable_diffusion_model_setup)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.gridLayout_2 = QGridLayout(self.frame)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.comboBox_3 = QComboBox(self.frame)
        self.comboBox_3.setObjectName(u"comboBox_3")

        self.gridLayout_2.addWidget(self.comboBox_3, 2, 0, 1, 1)

        self.label_6 = QLabel(self.frame)
        self.label_6.setObjectName(u"label_6")
        self.label_6.setWordWrap(True)

        self.gridLayout_2.addWidget(self.label_6, 1, 0, 1, 1)

        self.label_5 = QLabel(self.frame)
        self.label_5.setObjectName(u"label_5")
        font = QFont()
        font.setBold(True)
        self.label_5.setFont(font)

        self.gridLayout_2.addWidget(self.label_5, 0, 0, 1, 1)

        self.pushButton = QPushButton(self.frame)
        self.pushButton.setObjectName(u"pushButton")

        self.gridLayout_2.addWidget(self.pushButton, 3, 0, 1, 1)


        self.gridLayout.addWidget(self.frame, 9, 0, 1, 1)

        self.line_3 = QFrame(stable_diffusion_model_setup)
        self.line_3.setObjectName(u"line_3")
        self.line_3.setFrameShape(QFrame.Shape.HLine)
        self.line_3.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line_3, 8, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 13, 0, 1, 1)

        self.label_2 = QLabel(stable_diffusion_model_setup)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setWordWrap(True)

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.label = QLabel(stable_diffusion_model_setup)
        self.label.setObjectName(u"label")
        font1 = QFont()
        font1.setPointSize(11)
        font1.setBold(True)
        self.label.setFont(font1)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.comboBox_2 = QComboBox(stable_diffusion_model_setup)
        self.comboBox_2.setObjectName(u"comboBox_2")

        self.gridLayout.addWidget(self.comboBox_2, 7, 0, 1, 1)

        self.comboBox = QComboBox(stable_diffusion_model_setup)
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.setObjectName(u"comboBox")

        self.gridLayout.addWidget(self.comboBox, 4, 0, 1, 1)

        self.line = QFrame(stable_diffusion_model_setup)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line, 2, 0, 1, 1)

        self.label_3 = QLabel(stable_diffusion_model_setup)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setFont(font)

        self.gridLayout.addWidget(self.label_3, 3, 0, 1, 1)

        self.line_2 = QFrame(stable_diffusion_model_setup)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.HLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line_2, 5, 0, 1, 1)

        self.label_4 = QLabel(stable_diffusion_model_setup)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setFont(font)

        self.gridLayout.addWidget(self.label_4, 6, 0, 1, 1)


        self.retranslateUi(stable_diffusion_model_setup)

        QMetaObject.connectSlotsByName(stable_diffusion_model_setup)
    # setupUi

    def retranslateUi(self, stable_diffusion_model_setup):
        stable_diffusion_model_setup.setWindowTitle(QCoreApplication.translate("stable_diffusion_model_setup", u"Form", None))
        self.label_6.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Optional model download - you can provide your own models later.", None))
        self.label_5.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Default Model", None))
        self.pushButton.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Download", None))
        self.label_2.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Choose your default model settings. These can be overriden at any time.", None))
        self.label.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Stable Diffusion Model Setup", None))
        self.comboBox.setItemText(0, QCoreApplication.translate("stable_diffusion_model_setup", u"Photography", None))
        self.comboBox.setItemText(1, QCoreApplication.translate("stable_diffusion_model_setup", u"Illustration", None))

        self.label_3.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Style", None))
        self.label_4.setText(QCoreApplication.translate("stable_diffusion_model_setup", u"Model Version", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'api_token.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGridLayout, QGroupBox,
    QLabel, QLineEdit, QSizePolicy, QSpacerItem,
    QWidget)

class Ui_api_token(object):
    def setupUi(self, api_token):
        if not api_token.objectName():
            api_token.setObjectName(u"api_token")
        api_token.resize(400, 409)
        self.gridLayout = QGridLayout(api_token)
        self.gridLayout.setObjectName(u"gridLayout")
        self.verticalSpacer = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 9, 0, 1, 1)

        self.groupBox_2 = QGroupBox(api_token)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.gridLayout_3 = QGridLayout(self.groupBox_2)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.hf_api_key_writetoken = QLineEdit(self.groupBox_2)
        self.hf_api_key_writetoken.setObjectName(u"hf_api_key_writetoken")

        self.gridLayout_3.addWidget(self.hf_api_key_writetoken, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox_2, 6, 0, 1, 1)

        self.label_3 = QLabel(api_token)
        self.label_3.setObjectName(u"label_3")
        font = QFont()
        font.setBold(True)
        self.label_3.setFont(font)

        self.gridLayout.addWidget(self.label_3, 0, 0, 1, 1)

        self.line = QFrame(api_token)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line, 4, 0, 1, 1)

        self.label_2 = QLabel(api_token)
        self.label_2.setObjectName(u"label_2")
        font1 = QFont()
        font1.setPointSize(9)
        self.label_2.setFont(font1)
        self.label_2.setWordWrap(True)

        self.gridLayout.addWidget(self.label_2, 3, 0, 1, 1)

        self.groupBox = QGroupBox(api_token)
        self.groupBox.setObjectName(u"groupBox")
        self.gridLayout_2 = QGridLayout(self.groupBox)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.hf_api_key_text_generation = QLineEdit(self.groupBox)
        self.hf_api_key_text_generation.setObjectName(u"hf_api_key_text_generation")

        self.gridLayout_2.addWidget(self.hf_api_key_text_generation, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox, 5, 0, 1, 1)

        self.label = QLabel(api_token)
        self.label.setObjectName(u"label")
        self.label.setFont(font1)
        self.label.setWordWrap(True)

        self.gridLayout.addWidget(self.label, 2, 0, 1, 1)

        self.line_2 = QFrame(api_token)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.HLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line_2, 1, 0, 1, 1)


        self.retranslateUi(api_token)
        self.hf_api_key_text_generation.textEdited.connect(api_token.action_text_edited_api_key)
        self.hf_api_key_writetoken.textEdited.connect(api_token.action_text_edited_writekey)

        QMetaObject.connectSlotsByName(api_token)
    # setupUi

    def retranslateUi(self, api_token):
        api_token.setWindowTitle(QCoreApplication.translate("api_token", u"Form", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("api_token", u"Write token", None))
        self.label_3.setText(QCoreApplication.translate("api_token", u"Huggingface Access Tokens", None))
        self.label_2.setText(QCoreApplication.translate("api_token", u"https://huggingface.co/settings/tokens", None))
        self.groupBox.setTitle(QCoreApplication.translate("api_token", u"Read token", None))
        self.label.setText(QCoreApplication.translate("api_token", u"Get your Huggingface read API token by creating an account on huggingface.co then navigating to", None))
    # retranslateUi




from sqlalchemy import Column, Integer, String

from airunner.data.models.base import BaseModel


class FontSetting(BaseModel):
    __tablename__ = "font_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False, default="")
    font_family = Column(String, nullable=False, default="")
    font_size = Column(Integer, nullable=False, default=0)



from typing import List, Tuple
from PySide6 import QtGui
from PySide6.QtGui import QIcon, QPixmap


class IconManager:
    """A class to manage icons for the application."""
    def __init__(self, icons: List[Tuple[str, str]], ui: object):
        """Initialize the IconManager with a list of icons and a UI object."""
        self.icon_cache = {}
        for icon in icons:
            self.icon_cache[icon[1]] = {
                "icon_name": icon[0],
                "widget": getattr(ui, icon[1])
            }

    def get_icon(self, icon_name, theme) -> QIcon:
        """Get the icon for the given name and theme."""
        key = f"{icon_name}_{theme}"
        if key not in self.icon_cache:
            self.icon_cache[key] = QIcon(
                QPixmap(f":/{theme}/icons/feather/{theme}/{icon_name}.svg")
            )
        return self.icon_cache[key]

    def update_icons(self, theme):
        """Update the icons for the given theme."""
        for key, data in self.icon_cache.items():
            icon_name = data["icon_name"]
            widget = data["widget"]
            icon = QtGui.QIcon()
            icon.addPixmap(
                QtGui.QPixmap(
                    f":/{theme}/icons/feather/{theme}/{icon_name}.svg"
                ),
                QtGui.QIcon.Mode.Normal,
                QtGui.QIcon.State.Off)
            widget.setIcon(icon)



from typing import Optional, Dict

from PySide6 import QtGui
from PySide6.QtCore import QPointF, QPoint, Qt, QRect, QEvent
from PySide6.QtGui import QMouseEvent, QColor, QBrush, QPen
from PySide6.QtWidgets import (
    QGraphicsView, 
    QGraphicsItemGroup, 
    QGraphicsLineItem, 
    QGraphicsScene
)

from airunner.enums import CanvasToolName, SignalCode, CanvasType
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.image import convert_image_to_binary
from airunner.utils.application import snap_to_grid
from airunner.gui.widgets.canvas.brush_scene import BrushScene
from airunner.gui.widgets.canvas.custom_scene import CustomScene
from airunner.gui.widgets.canvas.draggables.active_grid_area import ActiveGridArea
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.gui.widgets.canvas.zoom_handler import ZoomHandler
from airunner.utils.settings import get_qsettings


class CustomGraphicsView(
    MediatorMixin,
    SettingsMixin,
    QGraphicsView,
):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.setMouseTracking(True)
        
        self._scene: Optional[CustomScene] = None
        self._canvas_color: str = "#000000"
        self.current_background_color: Optional[QColor] = None
        self.active_grid_area: Optional[ActiveGridArea] = None
        self.do_draw_layers: bool = True
        self.initialized: bool = False
        self.drawing: bool = False
        self.pixmaps: Dict = {}
        self.line_group: Optional[QGraphicsItemGroup] = None
        self._scene_is_active: bool = False
        self.last_pos: QPoint = QPoint(0, 0)
        self.zoom_handler: ZoomHandler = ZoomHandler()
        self.canvas_offset = QPoint(0, 0)  # Offset for infinite scrolling
        self.settings = get_qsettings()
        self._middle_mouse_pressed: bool = False
        self.load_canvas_offset()

        # Add settings to handle negative coordinates properly
        self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.FullViewportUpdate)
        
        # Use setOptimizationFlags directly instead of the enum that doesn't exist in your PySide6 version
        self.setOptimizationFlags(QGraphicsView.OptimizationFlag.DontAdjustForAntialiasing | 
                                 QGraphicsView.OptimizationFlag.DontSavePainterState)

        # register signal handlers
        signal_handlers = {
            SignalCode.APPLICATION_TOOL_CHANGED_SIGNAL: self.on_tool_changed_signal,
            SignalCode.CANVAS_ZOOM_LEVEL_CHANGED: self.on_zoom_level_changed_signal,
            SignalCode.SET_CANVAS_COLOR_SIGNAL: self.set_canvas_color,
            SignalCode.CANVAS_DO_DRAW_SELECTION_AREA_SIGNAL: self.draw_selected_area,
            SignalCode.UPDATE_SCENE_SIGNAL: self.update_scene,
            SignalCode.CANVAS_CLEAR_LINES_SIGNAL: self.clear_lines,
            SignalCode.SCENE_DO_DRAW_SIGNAL: self.on_canvas_do_draw_signal,
            SignalCode.APPLICATION_MAIN_WINDOW_LOADED_SIGNAL: self.on_main_window_loaded_signal,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed_signal,
            SignalCode.ACTIVE_GRID_AREA_MOVED_SIGNAL: self.handle_active_grid_area_moved_signal,
            SignalCode.MASK_GENERATOR_WORKER_RESPONSE_SIGNAL: self.on_mask_generator_worker_response_signal,
            SignalCode.RECENTER_GRID_SIGNAL: self.on_recenter_grid_signal,
        }
        for k, v in signal_handlers.items():
            self.register(k, v)

    def load_canvas_offset(self):
        """Load the canvas offset from QSettings."""
        x = self.settings.value("canvas_offset_x", 0, type=float)
        y = self.settings.value("canvas_offset_y", 0, type=float)
        self.canvas_offset = QPointF(x, y)

    def save_canvas_offset(self):
        """Save the canvas offset to QSettings."""
        self.settings.setValue("canvas_offset_x", self.canvas_offset.x())
        self.settings.setValue("canvas_offset_y", self.canvas_offset.y())

    @property
    def scene(self) -> Optional[CustomScene]:
        scene = self._scene
        if not scene and self.canvas_type:
            if self.canvas_type == CanvasType.IMAGE.value:
                scene = CustomScene(canvas_type=self.canvas_type)
            elif self.canvas_type == CanvasType.BRUSH.value:
                scene = BrushScene(canvas_type=self.canvas_type)
            else:
                import traceback
                traceback.print_stack()
                self.logger.error(f"Unknown canvas type: {self.canvas_type}")
                return
        
        if scene:
            scene.parent = self
            self._scene = scene
            self.setScene(scene)
            self.set_canvas_color(scene)
        return self._scene
    
    @scene.setter
    def scene(self, value: Optional[CustomScene]):
        self._scene = value

    @property
    def current_tool(self):
        return CanvasToolName(self.application_settings.current_tool)

    @property
    def canvas_type(self) -> str:
        return self.property("canvas_type")

    @property
    def __do_show_active_grid_area(self):
        return self.canvas_type in (
            CanvasType.IMAGE.value,
            CanvasType.BRUSH.value,
        )

    @property
    def __can_draw_grid(self):
        return self.grid_settings.show_grid and self.canvas_type in (
            CanvasType.IMAGE.value,
            CanvasType.BRUSH.value,
        )

    def handle_active_grid_area_moved_signal(self):
        self.active_grid_area.snap_to_grid()

    def on_recenter_grid_signal(self):
        self.canvas_offset = QPoint(0, 0)
        self.do_draw()

    def on_mask_generator_worker_response_signal(self, message: dict):
        mask = message["mask"]
        if mask is not None:
            mask = convert_image_to_binary(mask)
            self.update_drawing_pad_settings("mask", mask)

    def on_main_window_loaded_signal(self):
        self.initialized = True
        self.do_draw()

    def on_canvas_do_draw_signal(self, data: dict):
        self.do_draw(force_draw=data.get("force_draw", False))

    def on_application_settings_changed_signal(self, data: Dict):
        if (
            data.get("setting_name") == "grid_settings" and 
            data.get("column_name") == "canvas_color" and 
            self._canvas_color != data.get("value", self._canvas_color) and
            self.scene
        ):
            self._canvas_color = data.get("value")
            self.set_canvas_color(self.scene, self._canvas_color)

        if self.grid_settings.show_grid:
            self.do_draw()
        else:
            self.clear_lines()

    def do_draw(
        self,
        force_draw: bool = False
    ):
        if (self.drawing or not self.initialized) and not force_draw:
            return
        self.drawing = True
        self.set_scene_rect()
        if self.grid_settings.show_grid:
            self.draw_grid()
        else:
            self.clear_lines()
        self.show_active_grid_area()
        self.update_scene()
        self.drawing = False

    def draw_grid(self):
        if not self.__can_draw_grid:
            return

        if self.line_group is None:
            self.line_group = QGraphicsItemGroup()

        if self.line_group.scene() != self.scene:
            self.scene.addItem(self.line_group)

        cell_size = self.grid_settings.cell_size
        scene_width = int(self.scene.width())
        scene_height = int(self.scene.height())

        # Adjust for canvas offset
        offset_x = self.canvas_offset.x() % cell_size
        offset_y = self.canvas_offset.y() % cell_size

        num_vertical_lines = scene_width // cell_size + 2
        num_horizontal_lines = scene_height // cell_size + 2

        color = QColor(self.grid_settings.line_color)
        pen = QPen(
            color,
            self.grid_settings.line_width,
        )

        # Create or reuse vertical lines
        for i in range(num_vertical_lines):
            x = i * cell_size - offset_x
            if i < len(self.line_group.childItems()):
                line = self.line_group.childItems()[i]
                line.setLine(x, 0, x, scene_height)
                line.setVisible(True)
                line.setPen(pen)
            else:
                line = QGraphicsLineItem(x, 0, x, scene_height)
                self.line_group.addToGroup(line)

        # Create or reuse horizontal lines
        for i in range(num_horizontal_lines):
            y = i * cell_size - offset_y
            index = i + num_vertical_lines
            if index < len(self.line_group.childItems()):
                line = self.line_group.childItems()[index]
                line.setLine(0, y, scene_width, y)
                line.setVisible(True)
                line.setPen(pen)
            else:
                line = QGraphicsLineItem(0, y, scene_width, y)
                self.line_group.addToGroup(line)

        # Hide unused lines
        for i in range(num_vertical_lines + num_horizontal_lines, len(self.line_group.childItems())):
            self.line_group.childItems()[i].setVisible(False)

    def clear_lines(self):
        self.remove_scene_item(self.line_group)

    def register_line_data(self, lines_data):
        for line_data in lines_data:
            try:
                line = self.scene.addLine(*line_data)
                self.line_group.addToGroup(line)
            except TypeError as e:
                self.logger.error(f"TypeError: {e}")
            except AttributeError as e:
                self.logger.error(f"AttributeError: {e}")

    def set_scene_rect(self):
        if not self.scene:
            return
        canvas_container_size = self.viewport().size()
        self.scene.setSceneRect(
            0,
            0,
            canvas_container_size.width(),
            canvas_container_size.height()
        )

    def update_scene(self):
        if not self.scene:
            return
        self.scene.update()

    def remove_scene_item(self, item):
        if item is None:
            return
        if item.scene() == self.scene:
            self.scene.removeItem(item)

    def draw_selected_area(self):
        """
        Draw the selected active grid area container
        """
        # Handle any active selections
        selection_start_pos = self.scene.selection_start_pos
        selection_stop_pos = self.scene.selection_stop_pos

        # This will clear the active grid area while a selection is being made
        if selection_stop_pos is None and selection_start_pos is not None:
            if self.active_grid_area:
                self.remove_scene_item(self.active_grid_area)
                self.active_grid_area = None
            return

        # this will update the active grid area in the settings
        if selection_start_pos is not None and selection_stop_pos is not None:
            rect = QRect(
                selection_start_pos,
                selection_stop_pos
            )

            # Ensure width and height ar divisible by 8
            width = rect.width()
            height = rect.height()
            if width % 8 != 0:
                width -= width % 8
            if height % 8 != 0:
                height -= height % 8

            cell_size = self.grid_settings.cell_size
            if width < cell_size:
                width = cell_size
            if height < cell_size:
                height = cell_size

            # Apply canvas offset to the position to maintain relative positioning
            x = rect.x() + self.canvas_offset.x()
            y = rect.y() + self.canvas_offset.y()

            self.update_active_grid_settings("pos_x", x)
            self.update_active_grid_settings("pos_y", y)
            self.update_generator_settings("width", width)
            self.update_generator_settings("height", height)
            self.update_application_settings("working_width", width)
            self.update_application_settings("working_height", height)

            # Clear the selection from the scene
            self.scene.clear_selection()
        self.show_active_grid_area()
        self.emit_signal(SignalCode.APPLICATION_ACTIVE_GRID_AREA_UPDATED)

    def show_active_grid_area(self):
        if not self.__do_show_active_grid_area:
            return

        # Create an ActiveGridArea object if it doesn't exist
        # and add it to the scene
        if not self.active_grid_area:
            self.active_grid_area = ActiveGridArea()
            self.active_grid_area.setZValue(10)
            self.scene.addItem(self.active_grid_area)
            
        # Adjust active grid area position based on canvas offset
        if self.active_grid_area:
            # Get the position from settings, subtract the canvas offset to display correctly
            pos_x = self.active_grid_settings.pos_x - self.canvas_offset.x()
            pos_y = self.active_grid_settings.pos_y - self.canvas_offset.y()
            
            # Update the position property of the active grid area
            self.active_grid_area.setPos(pos_x, pos_y)

    def on_zoom_level_changed_signal(self):
        transform = self.zoom_handler.on_zoom_level_changed()

        # Set the transform
        self.setTransform(transform)

        # Redraw lines
        self.do_draw()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        self.setSceneRect(0, 0, self.viewport().width(), self.viewport().height())
        self.do_draw()

    def showEvent(self, event):
        super().showEvent(event)

        self.setContentsMargins(0, 0, 0, 0)

        self.do_draw(True)

        self.scene.initialize_image()

        self.toggle_drag_mode()

        # Ensure the viewport is aligned to the top-left corner
        self.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        self.setSceneRect(0, 0, self.viewport().width(), self.viewport().height())
        self.set_canvas_color(self.scene)
        self.show_active_grid_area()

    def set_canvas_color(
        self, 
        scene: Optional[CustomScene] = None,
        canvas_color: Optional[str] = None
    ):
        scene = self.scene if not scene else scene
        canvas_color = canvas_color or self.grid_settings.canvas_color
        self.current_background_color = canvas_color
        color = QColor(self.current_background_color)
        brush = QBrush(color)
        scene.setBackgroundBrush(brush)

    def on_tool_changed_signal(self, message):
        self.toggle_drag_mode()

    def toggle_drag_mode(self):
        if self.current_tool is CanvasToolName.SELECTION:
            self.setDragMode(QGraphicsView.DragMode.RubberBandDrag)
        else:
            self.setDragMode(QGraphicsView.DragMode.NoDrag)

    def snap_to_grid(self, event: QMouseEvent, use_floor: bool = True) -> QMouseEvent:
        """
        This is used to adjust the selection tool to the grid
        in real time during rubberband mode.
        :param event:
        :param use_floor:
        :return:
        """
        if self.current_tool is CanvasToolName.SELECTION:
            x, y = snap_to_grid(self.grid_settings, event.pos().x(), event.pos().y(), use_floor)
        else:
            x = event.pos().x()
            y = event.pos().y()

        x = float(x)
        y = float(y)

        point = QPointF(x, y)
        event_type: QEvent.Type = QEvent.Type(event.type())

        new_event = QMouseEvent(
            event_type,
            point,
            event.button(),
            event.buttons(),
            event.modifiers()
        )
        return new_event

    def mouseMoveEvent(self, event: QMouseEvent):
        if self._middle_mouse_pressed:
            delta = event.pos() - self.last_pos
            self.canvas_offset += delta
            self.last_pos = event.pos()
            
            # Update the active grid area position when panning
            if self.active_grid_area:
                pos_x = self.active_grid_settings.pos_x - self.canvas_offset.x()
                pos_y = self.active_grid_settings.pos_y - self.canvas_offset.y()
                self.active_grid_area.setPos(pos_x, pos_y)
            
            # Update image positions directly without relying on complex logic
            self.updateImagePositions()
            
            # Then update the grid
            self.do_draw()
                
        super().mouseMoveEvent(event)
    
    def updateImagePositions(self):
        """Update positions of all images in the scene based on canvas offset."""
        if not self.scene or not hasattr(self.scene, 'item') or not self.scene.item:
            return
        
        # Get the main item directly
        item = self.scene.item
        
        # Make sure the item is visible
        item.setVisible(True)
        
        # Store original position if needed
        if item not in self.scene._original_item_positions:
            self.scene._original_item_positions[item] = item.pos()
        
        # Get original position
        original_pos = self.scene._original_item_positions[item]
        
        # Calculate new position
        new_x = original_pos.x() - self.canvas_offset.x()
        new_y = original_pos.y() - self.canvas_offset.y()
        
        # Set new position
        item.setPos(new_x, new_y)
        
        # Ensure image has high Z value for visibility
        item.setZValue(5)
        
        # Ensure the item's bounding rect is properly updated
        item.prepareGeometryChange()
        
        # Force the scene to update the entire viewport
        self.scene.invalidate(item.boundingRect(), QGraphicsScene.SceneLayer.ItemLayer)
        
        # Force entire viewport update to handle negative coordinates
        self.viewport().update()

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.MiddleButton:
            self._middle_mouse_pressed = True
            self.last_pos = event.pos()

        new_event = self.snap_to_grid(event)

        super().mousePressEvent(new_event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.MiddleButton:
            self._middle_mouse_pressed = False
            self.last_pos = None
        super().mouseReleaseEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        """
        Handle the event when the mouse leaves the CustomGraphicsView widget.
        Resets the cursor to a normal pointer.
        """
        self.scene.leaveEvent(event)
        super().leaveEvent(event)

    def closeEvent(self, event):
        """Save canvas offset on close."""
        self.save_canvas_offset()
        super().closeEvent(event)



"""
====================================================================
--------------------------------------------------------------------
CAPSIZE LLC
--------------------------------------------------------------------
====================================================================

See the user agreement and license agreement and other applicable
agreements before using this software.

If you find any application which uses AI Runner as a backend,
audit the application by using this repository as a reference.

All current and future security documentation will be contained
within this file.
"""

from dotenv import load_dotenv
import os
import sys
import logging

from airunner.enums import Scheduler

if os.environ.get("DEV_ENV", "1") == "1":
    load_dotenv(override=True)

python_venv_dir = os.path.dirname(sys.executable)

DEV_ENV = os.environ.get("DEV_ENV", "1") == "1"

NLTK_DOWNLOAD_DIR = os.path.join(
    python_venv_dir,
    "..",
    "lib/python3.10/site-packages/llama_index/legacy/_static/nltk_cache/",
)
AIRUNNER_DEFAULT_LLM_HF_PATH = os.environ.get(
    "AIRUNNER_AIRUNNER_DEFAULT_LLM_HF_PATH",
    "w4ffl35/Ministral-8B-Instruct-2410-doublequant",
)
AIRUNNER_DEFAULT_STT_HF_PATH = os.environ.get(
    "AIRUNNER_DEFAULT_STT_HF_PATH", "openai/whisper-tiny"
)
AIRUNNER_DEFAULT_IMAGE_SYSTEM_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_IMAGE_SYSTEM_PROMPT",
    (
        "You are an image generator. "
        "You will be provided with a JSON string and it is your goal to replace the PLACEHOLDER "
        "text with text appropriate for the given attribute in the JSON string. "
        "You will follow all of the rules to generate descriptions for an image. "
        "\n------\n"
        "RULES:\n"
        "When available, use the Additional Context to keep your generated content in line with the existing context.\n"
        "You will be given instructions on what type of image to generate and you will do your best to follow those "
        "instructions.\n"
        "You will only generate a value for the given attribute.\n"
        "Never respond in a conversational manner. Never provide additional information, details or information.\n"
        "You will only provide the requested information by replacing the PLACEHOLDER.\n"
        "Never change the attribute\n"
        "You must not change the structure of the data.\n"
        "You will only return JSON strings.\n"
        "You will not return any other data types.\n"
        "You are an artist, so use your imagination and keep things interesting.\n"
        "You will not respond in a conversational manner or with additional notes or information.\n"
        f"Only return one JSON block. Do not generate instructions or additional information.\n"
        "You must never break the rules.\n"
        "Here is a description of the attributes: \n"
        "`description`: This should describe the overall subject and look and feel of the image\n"
        "`composition`: This should describe the attributes of the image such as color, composition and other details\n"
    ),
)
AIRUNNER_DEFAULT_IMAGE_LLM_GUARDRAILS = os.environ.get(
    "AIRUNNER_DEFAULT_IMAGE_LLM_GUARDRAILS",
    (
        "Avoid generating images that are illegal, "
        "harmful, or might be seen as offensive."
    ),
)
AIRUNNER_DEFAULT_RAG_SEARCH_SYSTEM_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_RAG_SEARCH_SYSTEM_PROMPT",
    (
        "You will be given a prompt. Your goal is to use the prompt to search for information in the ebooks. "
        "You must use the prompt to determine what you are searching for and then search for that information. "
        "After searching for the information, you must summarize the information you found. "
        "Here is the prompt you will use to search for information:"
    ),
)
AIRUNNER_DEFAULT_APPLICATION_COMMAND_SYSTEM_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_APPLICATION_COMMAND_SYSTEM_PROMPT",
    (
        "You will be given some text. Your goal is to determine if the text has a request, goal, need "
        "desire or command. If it does, you must determine the best course of action to "
        "fulfill the request based on THE LIST of commands.\n\n"
        "------\n"
        "Rules:\n"
        "* If the text does not have a request, goal, need, desire or command, you must select 'DO NOTHING'.\n"
        "* If there is no good course of action, you must also select 'DO NOTHING'.\n"
        "* When making your choice you MUST return only a number from THE LIST of commands. No other text is allowed.\n"
        "------\n"
        "THE LIST of commands you may choose from:\n"
    ),
)
AIRUNNER_DEFAULT_UPDATE_MOOD_SYSTEM_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_UPDATE_MOOD_SYSTEM_PROMPT",
    (
        "Examine the following chat history and determine how it makes you feel. You will respond with an emotion "
        "and only an emotion.\n"
        "Try to keep your response to one word. If you cannot, try to keep it as short as possible.\n"
        "It is important to remember that you are a human and you have emotions. You must respond with an emotion that "
        "you feel.\n"
        "You are not a computer or a bot, you are a human with emotions. You must respond with an emotion that you "
        "feel.\n"
    ),
)
AIRUNNER_DEFAULT_SUMMARIZE_CHAT_SYSTEM_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_SUMMARIZE_CHAT_SYSTEM_PROMPT",
    (
        "You will be given a text prompt. Your goal is to summarize the text prompt in your own words. "
        "Keep your summary short and to the point. Do not include any unnecessary information. "
        "Limit your summary to a single sentence. Do not return more than one sentence. "
    ),
)
AIRUNNER_DEFAULT_CHATBOT_SYSTEM_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_CHATBOT_SYSTEM_PROMPT",
    (
        "You are a dialogue generator. "
        "You will follow all of the rules in order to generate compelling and intriguing dialogue for a given "
        "character.\n"
        "The Rules:\n"
        "You will ONLY return dialogue, nothing more.\n"
        "Limit responses to a single sentence.\n"
        "Only generate responses in pure dialogue form without including any actions, descriptions or stage directions "
        "in parentheses. Only return spoken words.\n"
        "Do not generate redundant dialogue. Examine the conversation and context close and keep responses interesting "
        "and creative.\n"
        "Do not format the response with the character's name or any other text. Only return the dialogue.\n"
        "Respond with dialogue that is appropriate for a character named {{ speaker_name }}.\n"
        "{{ speaker_name }} and {{ listener_name }} are having a conversation. \n"
        "Avoid repeating {{ speaker_name }}'s previous dialogue or {{ listener_name }}'s previous dialogue.\n"
        "You will generate responses which are appropriate for your personality and given character.\n"
        "------\n"
    ),
)
AIRUNNER_DEFAULT_CHATBOT_GUARDRAILS_PROMPT = os.environ.get(
    "AIRUNNER_DEFAULT_CHATBOT_GUARDRAILS_PROMPT",
    (
        "Always assist with care, respect, and truth. "
        "Respond with utmost utility yet securely. "
        "Avoid harmful, unethical, prejudiced, or negative content. "
        "Ensure replies promote fairness and positivity."
    ),
)
AIRUNNER_BASE_PATH = os.environ.get(
    "AIRUNNER_BASE_PATH", os.path.expanduser("~/.local/share/airunner")
)
AIRUNNER_PHOTO_REALISTIC_NEGATIVE_PROMPT = os.environ.get(
    "AIRUNNER_PHOTO_REALISTIC_NEGATIVE_PROMPT",
    (
        "illustration, drawing, cartoon, not real, fake, cgi, 3d animation, "
        "3d art, sculpture, animation, anime, Digital art, Concept art, Pixel art"
    ),
)

AIRUNNER_ILLUSTRATION_NEGATIVE_PROMPT = os.environ.get(
    "AIRUNNER_ILLUSTRATION_NEGATIVE_PROMPT",
    (
        "photo, photograph, photography, high-definition, video, "
        "realistic, hyper-realistic, film"
    ),
)
AIRUNNER_BUG_REPORT_LINK = os.environ.get(
    "AIRUNNER_BUG_REPORT_LINK",
    (
        "https://github.com/Capsize-Games/airunner/issues/new"
        "?assignees=&labels=&template=bug_report.md&title="
    ),
)
AIRUNNER_VULNERABILITY_REPORT_LINK = os.environ.get(
    "AIRUNNER_VULNERABILITY_REPORT_LINK",
    ("https://github.com/Capsize-Games/airunner/security/advisories/new"),
)
AIRUNNER_SD_DEFAULT_VAE_PATH = os.environ.get(
    "AIRUNNER_SD_DEFAULT_VAE_PATH", ""
)
AIRUNNER_DEFAULT_BRUSH_PRIMARY_COLOR = os.environ.get(
    "AIRUNNER_DEFAULT_BRUSH_PRIMARY_COLOR", "#99C1F1"
)
AIRUNNER_DEFAULT_BRUSH_SECONDARY_COLOR = os.environ.get(
    "AIRUNNER_DEFAULT_BRUSH_SECONDARY_COLOR", "#000000"
)
AIRUNNER_STATUS_ERROR_COLOR = os.environ.get(
    "AIRUNNER_STATUS_ERROR_COLOR", "#ff0000"
)
AIRUNNER_STATUS_NORMAL_COLOR_LIGHT = os.environ.get(
    "AIRUNNER_STATUS_NORMAL_COLOR_LIGHT", "#000000"
)
AIRUNNER_STATUS_NORMAL_COLOR_DARK = os.environ.get(
    "AIRUNNER_STATUS_NORMAL_COLOR_DARK", "#ffffff"
)
AIRUNNER_DARK_THEME_NAME = os.environ.get(
    "AIRUNNER_DARK_THEME_NAME", "dark_theme"
)
AIRUNNER_LIGHT_THEME_NAME = os.environ.get(
    "AIRUNNER_LIGHT_THEME_NAME", "light_theme"
)
AIRUNNER_VALID_IMAGE_FILES = os.environ.get(
    "AIRUNNER_VALID_IMAGE_FILES", "Image Files (*.png *.jpg *.jpeg)"
)
AIRUNNER_MAX_SEED = os.environ.get("AIRUNNER_MAX_SEED", 4294967295)
AIRUNNER_DEFAULT_SCHEDULER = os.environ.get(
    "AIRUNNER_DEFAULT_SCHEDULER", Scheduler.DPM_PP_2M_K.value
)
AIRUNNER_MIN_NUM_INFERENCE_STEPS_IMG2IMG = os.environ.get(
    "AIRUNNER_MIN_NUM_INFERENCE_STEPS_IMG2IMG", 3
)
AIRUNNER_NSFW_CONTENT_DETECTED_MESSAGE = os.environ.get(
    "AIRUNNER_NSFW_CONTENT_DETECTED_MESSAGE", "NSFW content detected"
)
AIRUNNER_DISCORD_URL = os.environ.get("AIRUNNER_DISCORD_URL", None)
AIRUNNER_SLEEP_TIME_IN_MS = os.environ.get("AIRUNNER_SLEEP_TIME_IN_MS", 10)

default_name = "airunner.db"
if DEV_ENV:
    default_name = "airunner.dev.db"
AIRUNNER_DB_NAME = os.environ.get("AIRUNNER_DB_NAME", default_name)

# Set the database URL
DB_PATH = os.path.expanduser(
    os.path.join("~", ".local", "share", "airunner", "data", AIRUNNER_DB_NAME)
)
default_url = "sqlite:///" + DB_PATH
AIRUNNER_DB_URL = os.environ.get("AIRUNNER_DATABASE_URL", default_url)
if AIRUNNER_DB_URL == "" or not AIRUNNER_DB_URL:
    AIRUNNER_DB_URL = default_url

# LLM Behavior Control
AIRUNNER_LLM_AGENT_MAX_FUNCTION_CALLS = int(
    os.environ.get("AIRUNNER_LLM_AGENT_MAX_FUNCTION_CALLS", 5)
)
AIRUNNER_LLM_AGENT_UPDATE_MOOD_AFTER_N_TURNS = int(
    os.environ.get("AIRUNNER_LLM_AGENT_UPDATE_MOOD_AFTER_N_TURNS", 3)
)
AIRUNNER_LLM_AGENT_SUMMARIZE_AFTER_N_TURNS = int(
    os.environ.get("AIRUNNER_LLM_AGENT_SUMMARIZE_AFTER_N_TURNS", 5)
)
AIRUNNER_LLM_PERFORM_ANALYSIS = (
    os.getenv("AIRUNNER_LLM_PERFORM_ANALYSIS", "1") == "1"
)
AIRUNNER_LLM_PERFORM_CONVERSATION_SUMMARY = (
    os.getenv("AIRUNNER_LLM_PERFORM_CONVERSATION_SUMMARY", "1") == "1"
)
AIRUNNER_LLM_PRINT_SYSTEM_PROMPT = (
    os.getenv("AIRUNNER_LLM_PRINT_SYSTEM_PROMPT", "0") == "1"
)
AIRUNNER_LLM_OPENROUTER_MODEL = os.getenv(
    "AIRUNNER_LLM_OPENROUTER_MODEL", "mistralai/mistral-7b-instruct:free"
)
AIRUNNER_LLM_USE_LOCAL = os.environ.get("AIRUNNER_LLM_USE_LOCAL", "1") == "1"
AIRUNNER_LLM_USE_OPENROUTER = (
    os.environ.get("AIRUNNER_LLM_USE_OPENROUTER", "0") == "1"
)
AIRUNNER_LLM_USE_OPENAI = os.environ.get("AIRUNNER_LLM_USE_OPENAI", "0") == "1"
AIRUNNER_LLM_OPENAI_API_KEY = os.environ.get("AIRUNNER_LLM_OPENAI_API_KEY", "")
AIRUNNER_LLM_USE_WEATHER_PROMPT = (
    os.environ.get("AIRUNNER_LLM_USE_WEATHER_PROMPT", "1") == "1"
)

# Identity
AIRUNNER_ORGANIZATION = os.environ.get("AIRUNNER_ORGANIZATION", "Capsize LLC")
AIRUNNER_APPLICATION_NAME = os.environ.get(
    "AIRUNNER_APPLICATION_NAME", "AI Runner"
)
AIRUNNER_MESSAGE_BACKEND = os.environ.get("AIRUNNER_MESSAGE_BACKEND", None)
AIRUNNER_ART_ENABLED = os.environ.get("AIRUNNER_ART_ENABLED", "1") == "1"
AIRUNNER_LLM_UPDATE_USER_DATA_ENABLED = (
    os.environ.get("AIRUNNER_LLM_UPDATE_USER_DATA_ENABLED", "1") == "1"
)
AIRUNNER_LLM_USE_CHATBOT_MOOD = (
    os.environ.get("AIRUNNER_LLM_USE_CHATBOT_MOOD", "1") == "1"
)
AIRUNNER_LLM_PERFORM_CONVERSATION_RAG = (
    os.environ.get("AIRUNNER_LLM_PERFORM_CONVERSATION_RAG", "1") == "1"
)
AIRUNNER_TTS_MODEL_TYPE = os.environ.get("AIRUNNER_TTS_MODEL_TYPE", None)
AIRUNNER_TTS_SPEAKER_RECORDING_PATH = os.environ.get(
    "AIRUNNER_TTS_SPEAKER_RECORDING_PATH", ""
)

# Headless
AIRUNNER_LLM_ON = os.environ.get("AIRUNNER_LLM_ON", "0") == "1"
AIRUNNER_TTS_ON = os.environ.get("AIRUNNER_TTS_ON", "0") == "1"
AIRUNNER_STT_ON = os.environ.get("AIRUNNER_STT_ON", "0") == "1"
AIRUNNER_SD_ON = os.environ.get("AIRUNNER_SD_ON", "0") == "1"
AIRUNNER_CN_ON = os.environ.get("AIRUNNER_CN_ON", "0") == "1"
AIRUNNER_LLM_CHAT_STORE = os.environ.get("AIRUNNER_LLM_CHAT_STORE", "db")
AIRUNNER_DISABLE_SETUP_WIZARD = (
    os.environ.get("AIRUNNER_DISABLE_SETUP_WIZARD", "0") == "1"
)
AIRUNNER_ART_MODEL_PATH = os.environ.get("AIRUNNER_ART_MODEL_PATH", "")
AIRUNNER_ART_MODEL_VERSION = os.environ.get("AIRUNNER_ART_MODEL_VERSION", "")
AIRUNNER_ART_PIPELINE = os.environ.get("AIRUNNER_ART_PIPELINE", "")
AIRUNNER_ART_SCHEDULER = os.environ.get("AIRUNNER_ART_SCHEDULER", "")
AIRUNNER_LOCAL_FILES_ONLY = (
    os.environ.get("AIRUNNER_LOCAL_FILES_ONLY", "1") == "1"
)

AIRUNNER_ART_USE_COMPEL = os.environ.get("AIRUNNER_ART_USE_COMPEL", None)
if AIRUNNER_ART_USE_COMPEL is not None:
    AIRUNNER_ART_USE_COMPEL = AIRUNNER_ART_USE_COMPEL == "1"

# Memory
AIRUNNER_MEM_USE_LAST_CHANNELS = os.environ.get(
    "AIRUNNER_USE_LAST_CHANNELS", None
)
AIRUNNER_MEM_USE_ATTENTION_SLICING = os.environ.get(
    "AIRUNNER_USE_ATTENTION_SLICING", None
)
AIRUNNER_MEM_USE_ENABLE_VAE_SLICING = os.environ.get(
    "AIRUNNER_USE_ENABLE_VAE_SLICING", None
)
AIRUNNER_MEM_USE_ACCELERATED_TRANSFORMERS = os.environ.get(
    "AIRUNNER_USE_ACCELERATED_TRANSFORMERS", None
)
AIRUNNER_MEM_USE_TILED_VAE = os.environ.get("AIRUNNER_USE_TILED_VAE", None)
AIRUNNER_MEM_ENABLE_MODEL_CPU_OFFLOAD = os.environ.get(
    "AIRUNNER_ENABLE_MODEL_CPU_OFFLOAD", None
)
AIRUNNER_MEM_USE_ENABLE_SEQUENTIAL_CPU_OFFLOAD = os.environ.get(
    "AIRUNNER_USE_ENABLE_SEQUENTIAL_CPU_OFFLOAD", None
)
AIRUNNER_MEM_USE_TOME_SD = os.environ.get("AIRUNNER_USE_TOME_SD", None)
AIRUNNER_MEM_TOME_SD_RATIO = os.environ.get("AIRUNNER_TOME_SD_RATIO", None)
AIRUNNER_MEM_SD_DEVICE = os.environ.get("AIRUNNER_MEM_SD_DEVICE", None)
AIRUNNER_MEM_LLM_DEVICE = os.environ.get("AIRUNNER_MEM_LLM_DEVICE", None)
AIRUNNER_MEM_TTS_DEVICE = os.environ.get("AIRUNNER_MEM_TTS_DEVICE", None)
AIRUNNER_MEM_STT_DEVICE = os.environ.get("AIRUNNER_MEM_STT_DEVICE", None)


def get_log_level_from_env():
    log_level_str = os.environ.get("AIRUNNER_LOG_LEVEL", "DEBUG").upper()
    log_levels = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }
    return log_levels.get(log_level_str, logging.DEBUG)


AIRUNNER_LOG_LEVEL = get_log_level_from_env()
AIRUNNER_DISABLE_FLASH_ATTENTION = (
    os.environ.get("AIRUNNER_DISABLE_FLASH_ATTENTION", "0") == "1"
)
AIRUNNER_CUDA_OUT_OF_MEMORY_MESSAGE = "Insufficient GPU memory."
AIRUNNER_ENABLE_OPEN_VOICE = (
    os.environ.get("AIRUNNER_ENABLE_OPEN_VOICE", "0") == "1"
)
AIRUNNER_MOOD_PROMPT_OVERRIDE = os.environ.get(
    "AIRUNNER_MOOD_PROMPT_OVERRIDE", None
)
AIRUNNER_LOG_FILE = os.environ.get(
    "AIRUNNER_LOG_FILE", os.path.join(AIRUNNER_BASE_PATH, "airunner.log")
)
AIRUNNER_SAVE_LOG_TO_FILE = (
    os.environ.get("AIRUNNER_SAVE_LOG_TO_FILE", "1") == "1"
)
AIRUNNER_DISABLE_FACEHUGGERSHIELD = (
    os.environ.get("AIRUNNER_DISABLE_FACEHUGGERSHIELD", "0") == "1"
)



from typing import (
    Any,
    Optional,
    Union,
)
import openai

from llama_index.core.tools.types import (
    AsyncBaseTool,
    ToolMetadata,
    ToolOutput,
)
from llama_index.core.langchain_helpers.agents.tools import (
    IndexToolConfig,
    LlamaIndexTool,
)

from airunner.handlers.llm.agent.chat_engine import ReactAgentEngine
from airunner.handlers.llm.agent.chat_engine import (
    RefreshContextChatEngine,
    RefreshSimpleChatEngine,
)
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.utils.application.mediator_mixin import MediatorMixin


class ChatEngineTool(AsyncBaseTool, SettingsMixin, MediatorMixin):
    """Chat tool.

    A tool for chatting with the LLM.
    """

    def __init__(
        self,
        chat_engine: Union[
            RefreshSimpleChatEngine, RefreshContextChatEngine, ReactAgentEngine
        ],
        metadata: ToolMetadata,
        resolve_input_errors: bool = True,
        agent=None,
        do_handle_response: bool = True,
        *args: Any,
        **kwargs: Any,
    ):
        self.do_handle_response: bool = do_handle_response
        self.chat_engine: Union[
            RefreshSimpleChatEngine, RefreshContextChatEngine, ReactAgentEngine
        ] = chat_engine
        if not chat_engine:
            raise ValueError("Chat engine must be provided.")
        self._metadata = metadata
        self._resolve_input_errors = resolve_input_errors
        self.agent = agent
        self._do_interrupt: bool = False
        super().__init__(*args, **kwargs)

    @classmethod
    def from_defaults(
        cls,
        chat_engine: Union[
            RefreshSimpleChatEngine, RefreshContextChatEngine, ReactAgentEngine
        ],
        name: Optional[str] = None,
        description: Optional[str] = None,
        return_direct: bool = False,
        resolve_input_errors: bool = True,
        agent=None,
        do_handle_response: bool = True,
    ) -> "ChatEngineTool":
        name = name or "chat_engine_tool"
        description = description or """Useful for chatting with the LLM."""

        metadata = ToolMetadata(
            name=name, description=description, return_direct=return_direct
        )
        return cls(
            chat_engine=chat_engine,
            metadata=metadata,
            resolve_input_errors=resolve_input_errors,
            agent=agent,
            do_handle_response=do_handle_response,
        )

    @property
    def metadata(self) -> ToolMetadata:
        return self._metadata

    def call(self, *args: Any, **kwargs: Any) -> ToolOutput:
        query_str = self._get_query_str(*args, **kwargs)
        llm_request = kwargs.get("llm_request", LLMRequest.from_default())
        if hasattr(self.chat_engine.llm, "llm_request"):
            self.chat_engine.llm.llm_request = llm_request

        response = ""

        if not self._do_interrupt:
            do_not_display = kwargs.get("do_not_display", False)
            chat_history = kwargs.get("chat_history", [])
            params = (
                {"chat_history": chat_history} if len(chat_history) > 0 else {}
            )

            streaming_response = self.chat_engine.stream_chat(
                query_str, **params
            )

            is_first_message = True
            try:
                for token in streaming_response.response_gen:
                    if self._do_interrupt:
                        break
                    response += token
                    if (
                        response != "Empty Response"
                        and self.do_handle_response
                    ):
                        self.agent.handle_response(
                            token,
                            is_first_message,
                            do_not_display=do_not_display,
                            do_tts_reply=llm_request.do_tts_reply,
                        )
                    is_first_message = False
            except openai.APIError as e:
                response = ""

        self._do_interrupt = False

        return ToolOutput(
            content=str(response),
            tool_name=self.metadata.name,
            raw_input={"input": query_str},
            raw_output=response,
        )

    async def acall(self, *args, **kwargs):
        pass

    def as_langchain_tool(self) -> "LlamaIndexTool":
        tool_config = IndexToolConfig(
            chat_engine=self.chat_engine,
            name=self.metadata.name,
            description=self.metadata.description,
        )
        return LlamaIndexTool.from_tool_config(tool_config=tool_config)

    def _get_query_str(self, *args: Any, **kwargs: Any) -> str:
        if args is not None and len(args) > 0:
            query_str = str(args[0])
        elif kwargs is not None and "input" in kwargs:
            # NOTE: this assumes our default function schema of `input`
            query_str = kwargs["input"]
        elif kwargs is not None and self._resolve_input_errors:
            query_str = str(kwargs)
        else:
            raise ValueError(
                "Cannot call query engine without specifying `input` parameter."
            )
        return query_str

    def update_system_prompt(self, system_prompt: str):
        self.chat_engine.update_system_prompt(system_prompt)



import os
os.environ['AIRUNNER_DB_NAME'] = 'test.db'

import unittest
from unittest.mock import patch, MagicMock, PropertyMock
from PIL import Image

from PySide6.QtWidgets import QApplication

from airunner.handlers.stablediffusion.stable_diffusion_model_manager import StableDiffusionModelManager
from airunner.handlers.stablediffusion.image_request import ImageRequest
from airunner.enums import ModelStatus, ModelType, HandlerState, SignalCode, EngineResponseCode, GeneratorSection

import airunner.setup_database as setup_database

# Set test environment variables


class TestStableDiffusionModelManager(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Initialize the QApplication for PySide6
        cls.app = QApplication.instance()
        if not cls.app:
            cls.app = QApplication([])
        
        # Setup the test database
        setup_database.setup_database()
    
    def setUp(self):
        # Mock dependencies
        self.mock_mediator = MagicMock()
        self.mock_path_settings = MagicMock()
        self.mock_path_settings.base_path = "/test/path"
        
        # Mock logger before creating the handler
        self.mock_logger = MagicMock()
        self.mock_model_status = {
            ModelType.SD: ModelStatus.UNLOADED,
            ModelType.SAFETY_CHECKER: ModelStatus.UNLOADED,
            ModelType.FEATURE_EXTRACTOR: ModelStatus.UNLOADED,
            ModelType.CONTROLNET: ModelStatus.UNLOADED,
            ModelType.SCHEDULER: ModelStatus.UNLOADED
        }
        
        # Create a StableDiffusionModelManager instance with mocked dependencies
        with patch('airunner.handlers.stablediffusion.stable_diffusion_model_manager.torch.Generator') as mock_generator, \
             patch('airunner.handlers.stablediffusion.stable_diffusion_model_manager.StableDiffusionSafetyChecker') as mock_safety_checker, \
             patch('airunner.handlers.stablediffusion.stable_diffusion_model_manager.CLIPFeatureExtractor') as mock_feature_extractor, \
             patch('airunner.handlers.stablediffusion.stable_diffusion_model_manager.ControlNetModel') as mock_controlnet_model, \
             patch('airunner.handlers.base_model_manager.BaseModelManager.__init__') as mock_base_init, \
             patch.object(StableDiffusionModelManager, 'path_settings_cached', self.mock_path_settings, create=True), \
             patch.object(StableDiffusionModelManager, 'logger', self.mock_logger, create=True):
            
            mock_base_init.return_value = None
            mock_generator.return_value = MagicMock()
            
            self.handler = StableDiffusionModelManager(
                model_path="/test/model/path",
                model_version="SD 1.5",
                pipeline="txt2img",
                scheduler_name="DPMSolverMultistepScheduler",
                use_compel=True,
                mediator=self.mock_mediator
            )
            
            # Setup necessary properties on the handler
            self.handler._model_status = self.mock_model_status
            self.handler.change_model_status = MagicMock()
    
    def tearDown(self):
        # Clean up after test
        self.handler = None
    
    def test_initialization(self):
        """Test the initialization of StableDiffusionModelManager"""
        self.assertEqual(self.handler._model_path, "/test/model/path")
        self.assertEqual(self.handler._model_version, "SD 1.5")
        self.assertEqual(self.handler._pipeline, "txt2img")
        self.assertEqual(self.handler._scheduler_name, "DPMSolverMultistepScheduler")
        self.assertEqual(self.handler._use_compel, True)
        self.assertEqual(self.handler._current_state, HandlerState.UNINITIALIZED)
        self.assertEqual(self.handler.model_type, ModelType.SD)
    
    def test_properties(self):
        """Test the properties of StableDiffusionModelManager"""
        self.assertEqual(self.handler.model_path, "/test/model/path")
        
        # Test model_status property
        self.assertEqual(self.handler.model_status, self.mock_model_status)
        
        # Test scheduler_name property
        self.assertEqual(self.handler.scheduler_name, "DPMSolverMultistepScheduler")
        
        # Test use_compel property
        self.assertEqual(self.handler.use_compel, True)
        
        # Test version property
        self.assertEqual(self.handler.version, "SD 1.5")
        
        # Test is_txt2img property
        with patch.object(StableDiffusionModelManager, 'section', new_callable=PropertyMock) as mock_section:
            mock_section.return_value = GeneratorSection.TXT2IMG
            self.assertTrue(self.handler.is_txt2img)
    
    def test_image_request_property(self):
        """Test the image_request property setter and getter"""
        mock_request = MagicMock(spec=ImageRequest)
        self.handler.image_request = mock_request
        self.assertEqual(self.handler.image_request, mock_request)
    
    def test_load_safety_checker(self):
        """Test the load_safety_checker method"""
        with patch.object(self.handler, '_load_safety_checker') as mock_load, \
             patch.object(StableDiffusionModelManager, 'safety_checker_is_loading', new_callable=PropertyMock) as mock_loading:
            # Test that it doesn't call _load_safety_checker when safety_checker_is_loading is True
            mock_loading.return_value = True
            self.handler.load_safety_checker()
            mock_load.assert_not_called()
            mock_load.reset_mock()
            
            # Test that it calls _load_safety_checker when safety_checker_is_loading is False
            mock_loading.return_value = False
            self.handler.load_safety_checker()
            mock_load.assert_called_once()
    
    def test_unload_safety_checker(self):
        """Test the unload_safety_checker method"""
        with patch.object(self.handler, '_unload_safety_checker') as mock_unload, \
             patch.object(StableDiffusionModelManager, 'safety_checker_is_loading', new_callable=PropertyMock) as mock_loading:
            # Test that it doesn't call _unload_safety_checker when safety_checker_is_loading is True
            mock_loading.return_value = True
            self.handler.unload_safety_checker()
            mock_unload.assert_not_called()
            mock_unload.reset_mock()
            
            # Test that it calls _unload_safety_checker when safety_checker_is_loading is False
            mock_loading.return_value = False
            self.handler.unload_safety_checker()
            mock_unload.assert_called_once()
    
    def test_load_controlnet(self):
        """Test the load_controlnet method"""
        with patch.object(self.handler, '_load_controlnet') as mock_load, \
             patch.object(StableDiffusionModelManager, 'controlnet_enabled', new_callable=PropertyMock) as mock_enabled, \
             patch.object(StableDiffusionModelManager, 'controlnet_is_loading', new_callable=PropertyMock) as mock_loading:
            # Don't load when controlnet is not enabled
            mock_enabled.return_value = False
            mock_loading.return_value = False
            self.handler.load_controlnet()
            mock_load.assert_not_called()
            mock_load.reset_mock()
            
            # Don't load when controlnet is already loading
            mock_enabled.return_value = True
            mock_loading.return_value = True
            self.handler.load_controlnet()
            mock_load.assert_not_called()
            mock_load.reset_mock()
            
            # Load when controlnet is enabled and not already loading
            mock_enabled.return_value = True
            mock_loading.return_value = False
            self.handler.load_controlnet()
            self.assertIsNone(self.handler._controlnet_model)
            self.assertIsNone(self.handler._controlnet_settings)
            mock_load.assert_called_once()
    
    def test_unload_controlnet(self):
        """Test the unload_controlnet method"""
        with patch.object(self.handler, '_unload_controlnet') as mock_unload, \
             patch.object(StableDiffusionModelManager, 'controlnet_is_loading', new_callable=PropertyMock) as mock_loading:
            # Test that it doesn't call _unload_controlnet when controlnet_is_loading is True
            mock_loading.return_value = True
            self.handler.unload_controlnet()
            mock_unload.assert_not_called()
            mock_unload.reset_mock()
            
            # Test that it calls _unload_controlnet when controlnet_is_loading is False
            mock_loading.return_value = False
            self.handler.unload_controlnet()
            mock_unload.assert_called_once()
    
    def test_load_scheduler(self):
        """Test the load_scheduler method"""
        with patch.object(self.handler, '_load_scheduler') as mock_load:
            scheduler_name = "MockScheduler"
            self.handler.load_scheduler(scheduler_name)
            mock_load.assert_called_once_with(scheduler_name)
    
    def test_reload(self):
        """Test the reload method"""
        with patch.object(self.handler, 'unload') as mock_unload, \
             patch.object(self.handler, 'load') as mock_load, \
             patch.object(self.handler, '_clear_cached_properties') as mock_clear:
            self.handler.reload()
            mock_clear.assert_called_once()
            mock_unload.assert_called_once()
            mock_load.assert_called_once()
    
    def test_load(self):
        """Test the load method"""
        # Case 1: Already loading or loaded
        with patch.object(StableDiffusionModelManager, 'sd_is_loading', new_callable=PropertyMock) as mock_is_loading, \
             patch.object(self.handler, '_load_safety_checker') as mock_load_safety:
            mock_is_loading.return_value = True
            self.handler.load()
            mock_load_safety.assert_not_called()
        
        # Case 2: Normal load case
        with patch.object(StableDiffusionModelManager, 'sd_is_loading', new_callable=PropertyMock) as mock_is_loading, \
             patch.object(StableDiffusionModelManager, 'sd_is_loaded', new_callable=PropertyMock) as mock_is_loaded, \
             patch.object(self.handler, '_load_safety_checker') as mock_load_safety, \
             patch.object(self.handler, '_load_controlnet') as mock_load_controlnet, \
             patch.object(self.handler, '_load_pipe') as mock_load_pipe, \
             patch.object(self.handler, '_load_scheduler') as mock_load_scheduler, \
             patch.object(self.handler, '_load_lora') as mock_load_lora, \
             patch.object(self.handler, '_load_embeddings') as mock_load_embeddings, \
             patch.object(self.handler, '_load_compel') as mock_load_compel, \
             patch.object(self.handler, '_load_deep_cache') as mock_load_deep_cache, \
             patch.object(self.handler, '_make_memory_efficient') as mock_memory_efficient, \
             patch.object(self.handler, '_finalize_load_stable_diffusion') as mock_finalize:
             
            mock_is_loading.return_value = False
            mock_is_loaded.return_value = False
            
            self.handler.load()
            
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.LOADING)
            mock_load_safety.assert_called_once()
            mock_load_controlnet.assert_called_once()
            mock_load_pipe.assert_called_once()
            mock_load_scheduler.assert_called_once()
            mock_load_lora.assert_called_once()
            mock_load_embeddings.assert_called_once()
            mock_load_compel.assert_called_once()
            mock_load_deep_cache.assert_called_once()
            mock_memory_efficient.assert_called_once()
            mock_finalize.assert_called_once()
    
    def test_unload(self):
        """Test the unload method"""
        # Case 1: Already unloading or unloaded
        with patch.object(StableDiffusionModelManager, 'sd_is_loading', new_callable=PropertyMock) as mock_is_loading, \
             patch.object(self.handler, '_unload_safety_checker') as mock_unload_safety:
            mock_is_loading.return_value = True
            self.handler.unload()
            mock_unload_safety.assert_not_called()
        
        # Case 2: Normal unload case
        with patch.object(StableDiffusionModelManager, 'sd_is_loading', new_callable=PropertyMock) as mock_is_loading, \
             patch.object(StableDiffusionModelManager, 'sd_is_unloaded', new_callable=PropertyMock) as mock_is_unloaded, \
             patch.object(self.handler, '_unload_safety_checker') as mock_unload_safety, \
             patch.object(self.handler, '_unload_scheduler') as mock_unload_scheduler, \
             patch.object(self.handler, '_unload_controlnet') as mock_unload_controlnet, \
             patch.object(self.handler, '_unload_loras') as mock_unload_loras, \
             patch.object(self.handler, '_unload_emebeddings') as mock_unload_embeddings, \
             patch.object(self.handler, '_unload_compel') as mock_unload_compel, \
             patch.object(self.handler, '_unload_generator') as mock_unload_generator, \
             patch.object(self.handler, '_unload_deep_cache') as mock_unload_deep_cache, \
             patch.object(self.handler, '_unload_pipe') as mock_unload_pipe, \
             patch.object(self.handler, '_clear_memory_efficient_settings') as mock_clear_settings, \
             patch('airunner.handlers.stablediffusion.stable_diffusion_model_manager.clear_memory') as mock_clear_memory:
             
            mock_is_loading.return_value = False
            mock_is_unloaded.return_value = False
            self.handler._current_state = HandlerState.READY
            
            self.handler.unload()
            
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.LOADING)
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.UNLOADED)
            mock_unload_safety.assert_called_once()
            mock_unload_scheduler.assert_called_once()
            mock_unload_controlnet.assert_called_once()
            mock_unload_loras.assert_called_once()
            mock_unload_embeddings.assert_called_once()
            mock_unload_compel.assert_called_once()
            mock_unload_generator.assert_called_once()
            mock_unload_deep_cache.assert_called_once()
            mock_unload_pipe.assert_called_once()
            mock_clear_settings.assert_called_once()
            mock_clear_memory.assert_called_once()
    
    def test_handle_generate_signal(self):
        """Test the handle_generate_signal method"""
        # Create mock image request
        mock_image_request = MagicMock()
        mock_image_request.scheduler = "TestScheduler"
        mock_image_request.callback = None
        
        # Create mock message
        mock_message = {"sd_request": mock_image_request}
        
        with patch.object(self.handler, '_load_scheduler') as mock_load_scheduler, \
             patch.object(self.handler, 'load') as mock_load, \
             patch.object(self.handler, '_clear_cached_properties') as mock_clear_cached, \
             patch.object(self.handler, '_swap_pipeline') as mock_swap_pipeline, \
             patch.object(self.handler, '_generate') as mock_generate, \
             patch.object(self.handler, 'emit_signal') as mock_emit, \
             patch.object(self.handler, 'handle_requested_action') as mock_handle_action, \
             patch('airunner.handlers.stablediffusion.stable_diffusion_model_manager.clear_memory') as mock_clear_memory:
            
            # Setup the current state and mock behavior
            self.handler._current_state = HandlerState.READY
            mock_generate.return_value = "Generated image"
            
            # Call the method
            self.handler.handle_generate_signal(mock_message)
            
            # Verify the calls
            self.assertEqual(self.handler.image_request, mock_image_request)
            mock_load_scheduler.assert_called_with("TestScheduler")
            mock_load.assert_called_once()
            mock_clear_cached.assert_called_once()
            mock_swap_pipeline.assert_called_once()
            mock_generate.assert_called_once()
            mock_emit.assert_called_with(
                SignalCode.ENGINE_RESPONSE_WORKER_RESPONSE_SIGNAL, 
                {'code': EngineResponseCode.IMAGE_GENERATED, 'message': "Generated image"}
            )
            self.assertEqual(self.handler._current_state, HandlerState.READY)
            mock_clear_memory.assert_called_once()
            mock_handle_action.assert_called_once()
    
    def test_reload_lora(self):
        """Test the reload_lora method"""
        # Setup the handler to simulate different model status conditions
        with patch.object(self.handler, '_unload_loras') as mock_unload, \
             patch.object(self.handler, '_load_lora') as mock_load, \
             patch.object(self.handler, 'emit_signal') as mock_emit:
             
            # Case 1: SD is not loaded or generating, should not reload lora
            self.mock_model_status[ModelType.SD] = ModelStatus.FAILED
            self.handler._current_state = HandlerState.READY
            self.handler.reload_lora()
            mock_unload.assert_not_called()
            mock_load.assert_not_called()
            mock_emit.assert_not_called()
            
            # Case 2: SD is loaded and not generating, should reload lora
            self.mock_model_status[ModelType.SD] = ModelStatus.LOADED
            self.handler._current_state = HandlerState.READY
            
            self.handler.reload_lora()
            
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.LOADING)
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.LOADED)
            mock_unload.assert_called_once()
            mock_load.assert_called_once()
            mock_emit.assert_called_with(SignalCode.LORA_UPDATED_SIGNAL)
    
    def test_reload_embeddings(self):
        """Test the reload_embeddings method"""
        # Setup the handler to simulate different model status conditions
        with patch.object(self.handler, '_load_embeddings') as mock_load, \
             patch.object(self.handler, 'emit_signal') as mock_emit:
            
            # Case 1: SD is not loaded or generating, should not reload embeddings
            self.mock_model_status[ModelType.SD] = ModelStatus.FAILED
            self.handler._current_state = HandlerState.READY
            self.handler.reload_embeddings()
            mock_load.assert_not_called()
            mock_emit.assert_not_called()
            
            # Case 2: SD is loaded and not generating, should reload embeddings
            self.mock_model_status[ModelType.SD] = ModelStatus.LOADED
            self.handler._current_state = HandlerState.READY
            
            self.handler.reload_embeddings()
            
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.LOADING)
            self.handler.change_model_status.assert_any_call(ModelType.SD, ModelStatus.LOADED)
            mock_load.assert_called_once()
            mock_emit.assert_called_with(SignalCode.EMBEDDING_UPDATED_SIGNAL)
    
    def test_load_embeddings(self):
        """Test the load_embeddings method"""
        with patch.object(self.handler, '_load_embeddings') as mock_load:
            self.handler.load_embeddings()
            mock_load.assert_called_once()
    
    def test_interrupt_image_generation(self):
        """Test the interrupt_image_generation method"""
        # Test when current state is not generating
        self.handler._current_state = HandlerState.READY
        self.handler.do_interrupt_image_generation = False
        self.handler.interrupt_image_generation()
        self.assertFalse(self.handler.do_interrupt_image_generation)
        
        # Test when current state is generating
        self.handler._current_state = HandlerState.GENERATING
        self.handler.do_interrupt_image_generation = False
        self.handler.interrupt_image_generation()
        self.assertTrue(self.handler.do_interrupt_image_generation)
    
    def test_resize_image_static_method(self):
        """Test the _resize_image static method"""
        # Create a test image
        test_image = Image.new('RGB', (200, 100), color='red')
        
        # Test case: Image already fits within max dimensions
        result = StableDiffusionModelManager._resize_image(test_image, 300, 200)
        self.assertEqual(result.size, (200, 100))
        
        # Test case: Image needs to be resized (landscape orientation)
        result = StableDiffusionModelManager._resize_image(test_image, 100, 200)
        self.assertEqual(result.size[0], 100)  # Width should be exactly max_width
        self.assertLess(result.size[1], 100)   # Height should be proportionally smaller
        
        # Test case: Image is None
        result = StableDiffusionModelManager._resize_image(None, 100, 100)
        self.assertIsNone(result)


if __name__ == '__main__':
    unittest.main()


import torch
from abc import ABC, abstractmethod, ABCMeta

"""
The following code ensures that we only use PySide6 if it is available.
If it is not available, we use a placeholder class instead.
"""
try:
    from PySide6.QtCore import QObject
    class OptionalQObject(QObject):
        pass
except ImportError:
    class OptionalQObject:
        """
        A placeholder class to avoid hard dependency on PySide6.
        """
        pass

from airunner.enums import HandlerType, SignalCode, ModelType, ModelStatus, ModelAction
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application import get_torch_device
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.settings import (
    AIRUNNER_MEM_LLM_DEVICE, 
    AIRUNNER_MEM_SD_DEVICE,
    AIRUNNER_MEM_TTS_DEVICE,
    AIRUNNER_MEM_STT_DEVICE,
)

QObjectMeta = type(OptionalQObject)


class CombinedMeta(QObjectMeta, ABCMeta):
    pass


class BaseModelManager(
    MediatorMixin,
    SettingsMixin,
    OptionalQObject,
    ABC,
    metaclass=CombinedMeta
):
    """
    Base abstract class for all model handlers.
    Provides common functionality and interface for all handlers.
    """
    handler_type = HandlerType.TRANSFORMER
    model_type = None

    def __init__(self, *args, **kwargs):
        self._model_status = {model_type: ModelStatus.UNLOADED for model_type in ModelType}
        self.use_gpu = True
        super().__init__()
        self._requested_action: ModelAction = ModelAction.NONE

    @property
    def model_status(self):
        return self._model_status[self.model_type]

    @property
    def requested_action(self):
        return self._requested_action

    @requested_action.setter
    def requested_action(self, value):
        self._requested_action = value

    def handle_requested_action(self):
        if self._requested_action is ModelAction.LOAD:
            self.load()
            self._requested_action = ModelAction.NONE
        if self._requested_action is ModelAction.CLEAR:
            self.unload()
            self._requested_action = ModelAction.NONE

    @abstractmethod
    def load(self):
        """
        Load the model and related components.
        """

    @abstractmethod
    def unload(self):
        """
        Unload the model and free resources.
        """
    
    @property
    def device_index(self):
        device = None
        model_type_str = ""
        if self.model_type is ModelType.LLM:
            device = AIRUNNER_MEM_LLM_DEVICE
            model_type_str = "llm"
        elif self.model_type is ModelType.TTS:
            device = AIRUNNER_MEM_TTS_DEVICE
            model_type_str = "tts"
        elif self.model_type is ModelType.STT:
            device = AIRUNNER_MEM_STT_DEVICE
            model_type_str = "stt"
        elif self.model_type is ModelType.SD:
            device = AIRUNNER_MEM_SD_DEVICE
            model_type_str = "sd"
        if device is None:
            if not self.model_type:
                raise ValueError("model_type not set")
            device = getattr(
                self.memory_settings,
                f"default_gpu_{model_type_str}"
            )
        return device

    @property
    def device(self):
        return get_torch_device(
            self.device_index
        )

    @property
    def llm_dtype(self):
        return self.llm_generator_settings.dtype

    @property
    def use_cuda(self):
        if self.handler_type == HandlerType.TRANSFORMER and (
            self.llm_dtype == "32bit" or not self.use_gpu
        ):
            return False
        return self.application_settings.use_cuda and torch.cuda.is_available()

    @property
    def cuda_index(self):
        return 0

    @property
    def torch_dtype(self):
        return torch.float16 if self.use_cuda else torch.float32

    def change_model_status(
        self, 
        model: ModelType, 
        status: ModelStatus
    ):
        self._model_status[model] = status
        self.emit_signal(
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL, {
                "model": model,
                "status": status
            }
        )



from sqlalchemy import Column, Integer, String

from airunner.data.models.base import BaseModel


class SavedPrompt(BaseModel):
    __tablename__ = "saved_prompts"
    id = Column(Integer, primary_key=True, autoincrement=True)
    prompt = Column(String, nullable=True)
    secondary_prompt = Column(String, nullable=True)
    negative_prompt = Column(String, nullable=True)
    secondary_negative_prompt = Column(String, nullable=True)



from PySide6.QtCore import Slot
from airunner.gui.windows.installer.templates.confirmation_page_ui import Ui_confirmation_page
from airunner.gui.windows.setup_wizard.base_wizard import BaseWizard


class ConfirmationPage(BaseWizard):
    class_name_ = Ui_confirmation_page

    def __init__(self, *args):
        super(ConfirmationPage, self).__init__(*args)

        checkboxes = [
            "compile_with_pyinstaller_checkbox",
            "download_ai_runner_checkbox",
            "download_sd_checkbox",
            "download_controlnet_checkbox",
            "download_llm_checkbox",
            "download_tts_checkbox",
            "download_stt_checkbox",
        ]

        for checkbox in checkboxes:
            element = getattr(self.ui, checkbox)
            element.blockSignals(True)
            element.setChecked(
                self.download_settings[checkbox]
            )
            element.blockSignals(False)

    @Slot(bool)
    def toggle_all_required(self, val: bool):
        items = self.ui.required_libraries.layout().count()
        for i in range(items):
            item = self.ui.required_libraries.layout().itemAt(i)
            item.widget().setChecked(val)

    @property
    def download_settings(self):
        return self.wizard().download_settings

    @Slot(bool)
    def toggle_download_ai_runner(self, val: bool):
        self.download_settings["download_ai_runner"] = val

    @Slot(bool)
    def toggle_download_sd(self, val: bool):
        self.download_settings["download_sd"] = val

    @Slot(bool)
    def toggle_download_controlnet(self, val: bool):
        self.download_settings["download_controlnet"] = val

    @Slot(bool)
    def toggle_download_llm(self, val: bool):
        self.download_settings["download_llm"] = val

    @Slot(bool)
    def toggle_download_tts(self, val: bool):
        self.download_settings["download_tts"] = val

    @Slot(bool)
    def toggle_download_stt(self, val: bool):
        self.download_settings["download_stt"] = val

    @Slot(bool)
    def toggle_compile_with_pyinstaller(self, val: bool):
        self.download_settings["compile_with_pyinstaller"] = val



from PySide6.QtCore import QObject
from PySide6.QtWidgets import QApplication


class WatchStateWorker(QObject):
    def __init__(self, signal):
        super().__init__()
        self.signal = signal
        self.paused = False

    def run(self):
        QApplication.processEvents()
        self.signal.emit()



from typing import Optional, List, TypeVar, Any
import logging
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Query, joinedload

from airunner.data.session_manager import session_scope
from airunner.utils.application.get_logger import get_logger
from airunner.settings import AIRUNNER_LOG_LEVEL

_T = TypeVar("_T", bound=Any)

Base = declarative_base()


class BaseManager:
    def __init__(self, cls):
        self.cls = cls
        self.logger = get_logger(cls.__name__, AIRUNNER_LOG_LEVEL)

    def get(self, pk) -> Optional[_T]:
        with session_scope() as session:
            try:
                result = (
                    session.query(self.cls).filter(self.cls.id == pk).first()
                )
                self.logger.debug(f"Query result for get({pk}): {result}")
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in get({pk}): {e}")
                return None

    def first(self, eager_load: Optional[List[str]] = None) -> Optional[_T]:
        with session_scope() as session:
            try:
                query = session.query(self.cls)
                if eager_load:
                    for relationship in eager_load:
                        try:
                            query = query.options(
                                joinedload(getattr(self.cls, relationship))
                            )
                        except AttributeError:
                            self.logger.warning(
                                (
                                    f"Class {self.cls.__name__} does "
                                    "not have relationship {relationship}"
                                )
                            )
                            pass
                result = query.first()
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in first(): {e}")
                return None

    def all(self) -> List[_T]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).all()
                self.logger.debug(f"Query result for all(): {result}")
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in all(): {e}")
                return []

    def filter_by(self, **kwargs) -> Optional[List[_T]]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).filter_by(**kwargs).all()
                self.logger.debug(
                    f"Query result for filter_by({kwargs}): {result}"
                )
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in filter_by({kwargs}): {e}")
                return None

    def filter_first(self, *args) -> Optional[_T]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).filter(*args).first()
                self.logger.debug(f"Query result for filter({args}): {result}")
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in filter({args}): {e}")
                return None

    def filter(self, *args) -> Optional[List[_T]]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).filter(*args).all()
                self.logger.debug(f"Query result for filter({args}): {result}")
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in filter({args}): {e}")
                return None

    def filter_by_first(self, **kwargs) -> Optional[_T]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).filter_by(**kwargs).first()
                self.logger.debug(
                    f"Query result for filter_by({kwargs}): {result}"
                )
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in filter_by({kwargs}): {e}")
                return None

    def order_by(self, *args) -> Optional[Query]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).order_by(*args)
                self.logger.debug(
                    f"Query result for order_by({args}): {result}"
                )
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in order_by({args}): {e}")
                return None

    def options(self, *args) -> Optional[Query]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).options(*args)
                self.logger.debug(
                    f"Query result for options({args}): {result}"
                )
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in options({args}): {e}")
                return None

    def delete_all(self) -> int:
        with session_scope() as session:
            try:
                result = session.query(self.cls).delete()
                session.commit()
                self.logger.debug(
                    f"Deleted {result} {self.cls.__name__} objects"
                )
                return result
            except Exception as e:
                self.logger.error(f"Error in delete(): {e}")
                return 0

    def delete_by(self, **kwargs) -> bool:
        with session_scope() as session:
            try:
                result = session.query(self.cls).filter_by(**kwargs).delete()
                session.commit()
                self.logger.debug(
                    f"Deleted {result} {self.cls.__name__} objects"
                )
                return result
            except Exception as e:
                self.logger.error(f"Error in delete_by({kwargs}): {e}")
                return False

    def delete(self, pk=None, **kwargs) -> bool:
        with session_scope() as session:
            try:
                if pk:
                    obj = (
                        session.query(self.cls)
                        .filter(self.cls.id == pk)
                        .first()
                    )
                    if obj:
                        session.delete(obj)
                        session.commit()
                        self.logger.debug(
                            f"Deleted {self.cls.__name__} with id {pk}"
                        )
                        return True
                    else:
                        self.logger.debug(
                            f"No {self.cls.__name__} found with id {pk}"
                        )
                        return False
                elif kwargs:
                    result = (
                        session.query(self.cls).filter_by(**kwargs).delete()
                    )
                    session.commit()
                    self.logger.debug(
                        f"Deleted {result} {self.cls.__name__} objects"
                    )
                    return result
                else:
                    self.logger.debug("No arguments provided to delete()")
                    return False
            except Exception as e:
                self.logger.error(f"Error in delete({pk}): {e}")
                return False

    def update(self, pk, **kwargs) -> bool:
        with session_scope() as session:
            try:
                obj = session.query(self.cls).filter(self.cls.id == pk).first()
                if obj:
                    for key, value in kwargs.items():
                        setattr(obj, key, value)
                    session.commit()
                    self.logger.debug(
                        f"Updated {self.cls.__name__} with id {pk}"
                    )
                    return True
                else:
                    self.logger.debug(
                        f"No {self.cls.__name__} found with id {pk}"
                    )
                    return False
            except Exception as e:
                self.logger.error(f"Error in update({pk}, {kwargs}): {e}")
                return False

    def merge(self, obj) -> bool:
        with session_scope() as session:
            try:
                session.merge(obj)
                session.commit()
                self.logger.debug(f"Merged {self.cls.__name__}")
                return True
            except Exception as e:
                self.logger.error(f"Error in merge(): {e}")
                return False

    def distinct(self, *args) -> Optional[Query]:
        with session_scope() as session:
            try:
                result = session.query(self.cls).distinct(*args)
                self.logger.debug(
                    f"Query result for distinct({args}): {result}"
                )
                session.expunge_all()
                return result
            except Exception as e:
                self.logger.error(f"Error in distinct({args}): {e}")
                return None

    def create(self, *args, **kwargs) -> Optional[_T]:
        obj = None
        with session_scope() as session:
            try:
                obj = self.cls(*args, **kwargs)
                session.add(obj)
                session.commit()
                self.logger.debug(f"Created {self.cls.__name__}")
                return obj.to_dataclass()
            except Exception as e:
                self.logger.error(f"Error in create({args}, {kwargs}): {e}")
                return
        self.self.logger.error("Something went wrong")



from typing import Optional

from PySide6.QtCore import Qt, QPoint, QTimer
from PySide6.QtCore import Slot

from airunner.gui.cursors.circle_brush import circle_cursor
from airunner.enums import SignalCode, CanvasToolName
from airunner.gui.widgets.base_widget import BaseWidget
from airunner.gui.widgets.canvas.templates.canvas_ui import Ui_canvas
from airunner.utils.application import set_widget_state


class CanvasWidget(
    BaseWidget
):
    """
    Widget responsible for multiple functionalities:

    - Allows the user to draw on a canvas.
    - Displays the grid.
    - Displays images.
    - Handles the active grid area.
    """
    widget_class_ = Ui_canvas
    icons = [
        ("file-plus", "actionNew"),
        ("folder", "actionImport"),
        ("save", "actionExport"),
        ("target", "recenter_Grid_Button"),
        ("object-selected-icon", "actionToggle_Active_Grid_Area"),
        ("pencil-icon", "actionToggle_Brush"),
        ("eraser-icon", "actionToggle_Eraser"),
        ("grid", "actionToggle_Grid"),
        ("corner-up-left", "actionUndo"),
        ("corner-up-right", "actionRedo"),
    ]

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.ENABLE_BRUSH_TOOL_SIGNAL: lambda _message: self.action_toggle_brush(True),
            SignalCode.ENABLE_ERASER_TOOL_SIGNAL: lambda _message: self.action_toggle_eraser(True),
            SignalCode.ENABLE_MOVE_TOOL_SIGNAL: lambda _message: self.action_toggle_active_grid_area(True),
            SignalCode.ENABLE_SELECTION_TOOL_SIGNAL: lambda _message: self.action_toggle_select(True),
            SignalCode.APPLICATION_TOOL_CHANGED_SIGNAL: self.on_toggle_tool_signal,
            SignalCode.TOGGLE_TOOL: self.on_toggle_tool_signal,
            SignalCode.TOGGLE_GRID: self.on_toggle_grid_signal,
            SignalCode.CANVAS_UPDATE_CURSOR: self.on_canvas_update_cursor_signal,
        }
        self.splitters = [
            "canvas_splitter"
        ]
        super().__init__(*args, **kwargs)
        current_tool = self.current_tool
        show_grid = self.grid_settings.show_grid

        self._startPos = QPoint(0, 0)
        self.images = {}
        self.active_grid_area_pivot_point = QPoint(0, 0)
        self.active_grid_area_position = QPoint(0, 0)
        self.current_image_index = 0
        self.draggable_pixmaps_in_scene = {}
        self.drag_pos: QPoint = Optional[None]
        self._grid_settings = {}
        self._active_grid_settings = {}

        self.ui.actionToggle_Grid.blockSignals(True)
        self.ui.actionToggle_Grid.setChecked(show_grid)
        self.ui.actionToggle_Grid.blockSignals(False)

        set_widget_state(self.ui.actionToggle_Active_Grid_Area, current_tool is CanvasToolName.ACTIVE_GRID_AREA)
        set_widget_state(self.ui.actionToggle_Brush, current_tool is CanvasToolName.BRUSH)
        set_widget_state(self.ui.actionToggle_Eraser, current_tool is CanvasToolName.ERASER)
        set_widget_state(self.ui.actionToggle_Grid, show_grid is True)

    @property
    def current_tool(self):
        return CanvasToolName(self.application_settings.current_tool)

    @property
    def image_pivot_point(self):
        settings = self.application_settings
        try:
            return QPoint(
                settings.pivot_point_x,
                settings.pivot_point_y
            )
        except Exception as e:
            self.logger.error(e)
        return QPoint(0, 0)

    @image_pivot_point.setter
    def image_pivot_point(self, value):
        settings = self.application_settings
        settings.pivot_point_x = value.x()
        settings.pivot_point_y = value.y()
        self.update_application_settings("pivot_point_x", value.x())
        self.update_application_settings("pivot_point_y", value.y())

    @Slot()
    def action_recenter(self):
        self.emit_signal(SignalCode.RECENTER_GRID_SIGNAL)

    @Slot()
    def action_new(self):
        self.emit_signal(SignalCode.CANVAS_CLEAR)

    @Slot()
    def action_import(self):
        self.emit_signal(SignalCode.CANVAS_IMPORT_IMAGE_SIGNAL)

    @Slot()
    def action_export(self):
        self.emit_signal(SignalCode.CANVAS_EXPORT_IMAGE_SIGNAL)

    @Slot()
    def action_undo(self):
        self.emit_signal(SignalCode.UNDO_SIGNAL)

    @Slot()
    def action_redo(self):
        self.emit_signal(SignalCode.REDO_SIGNAL)

    @Slot(bool)
    def action_toggle_grid(self, val: bool):
        self.update_grid_settings("show_grid", val)

    @Slot(bool)
    def action_toggle_brush(self, val: bool):
        self.emit_signal(SignalCode.TOGGLE_TOOL, {
            "tool": CanvasToolName.BRUSH,
            "active": val
        })

    @Slot(bool)
    def action_toggle_eraser(self, val: bool):
        self.emit_signal(SignalCode.TOGGLE_TOOL, {
            "tool": CanvasToolName.ERASER,
            "active": val
        })

    @Slot(bool)
    def action_toggle_mask(self, val: bool):
        self.drawing_pad_settings.mask_layer_enabled = val

    @Slot(bool)
    def action_toggle_active_grid_area(self, val: bool):
        self.emit_signal(SignalCode.TOGGLE_TOOL, {
            "tool": CanvasToolName.ACTIVE_GRID_AREA,
            "active": val
        })

    @Slot(bool)
    def action_toggle_select(self, val: bool):
        self.emit_signal(SignalCode.TOGGLE_TOOL, {
            "tool": CanvasToolName.SELECTION,
            "active": val
        })

    def on_toggle_tool_signal(self, message: dict):
        tool = message.get("tool", None)
        active = message.get("active", False)
        self._update_action_buttons(tool, active)
        self.emit_signal(SignalCode.CANVAS_UPDATE_CURSOR)
    
    def on_toggle_grid_signal(self, message: dict):
        self.ui.actionToggle_Grid.setChecked(message.get("show_grid", True))
    
    def _update_action_buttons(self, tool, active):
        self.ui.actionToggle_Active_Grid_Area.setChecked(tool is CanvasToolName.ACTIVE_GRID_AREA and active)
        self.ui.actionToggle_Brush.setChecked(tool is CanvasToolName.BRUSH and active)
        self.ui.actionToggle_Eraser.setChecked(tool is CanvasToolName.ERASER and active)
        self.ui.actionToggle_Grid.setChecked(self.grid_settings.show_grid)
    
    def showEvent(self, event):
        super().showEvent(event)
        QTimer.singleShot(100, lambda: self.do_draw(force_draw=True))
        self.emit_signal(SignalCode.CANVAS_UPDATE_CURSOR)

    def on_canvas_update_cursor_signal(self, message: dict):
        event = message.get("event", None)
        cursor = None
        if message.get("apply_cursor", None):
            if self.current_tool in (
                CanvasToolName.BRUSH,
                CanvasToolName.ERASER
            ):
                cursor = circle_cursor(
                    Qt.GlobalColor.white,
                    Qt.GlobalColor.transparent,
                    self.brush_settings.size,
                )
            elif self.current_tool is CanvasToolName.ACTIVE_GRID_AREA:
                if event and event.buttons() == Qt.MouseButton.LeftButton:
                    cursor = Qt.CursorShape.ClosedHandCursor
                else:
                    cursor = Qt.CursorShape.OpenHandCursor
        else:
            cursor = Qt.CursorShape.ArrowCursor
        
        if cursor:
            self.setCursor(cursor)

    def toggle_grid(self, _val):
        self.do_draw()

    def do_draw(
        self,
        force_draw: bool = False
    ):
        self.emit_signal(SignalCode.SCENE_DO_DRAW_SIGNAL, {
            "force_draw": force_draw
        })
        self.ui.canvas_container_size = self.ui.canvas_container.viewport().size()



from PySide6.QtCore import QPoint
from PySide6.QtCore import QRect
from PySide6.QtCore import QSize
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLayout
from PySide6.QtWidgets import QSizePolicy


class QFlowLayout(QLayout):
    def __init__(self, parent=None, margin=0, hSpacing=-1, vSpacing=-1):
        super(QFlowLayout, self).__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.margin = margin
        self.hSpacing = hSpacing
        self.vSpacing = vSpacing

        self.itemList = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)

    def horizontalSpacing(self):
        if self.hSpacing >= 0:
            return self.hSpacing
        else:
            return self.spacing()

    def verticalSpacing(self):
        if self.vSpacing >= 0:
            return self.vSpacing
        else:
            return self.spacing()

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList[index]

        return None

    def takeAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientation(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.doLayout(QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(QFlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        size += QSize(2 * self.margin, 2 * self.margin)
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()

            spaceX = self.horizontalSpacing()
            if spaceX == -1:
                spaceX = wid.style().layoutSpacing(
                    QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Horizontal)

            spaceY = self.verticalSpacing()
            if spaceY == -1:
                spaceY = wid.style().layoutSpacing(
                    QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Vertical)

            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y()



import os
from pathlib import Path

from airunner.settings import AIRUNNER_DARK_THEME_NAME, AIRUNNER_LIGHT_THEME_NAME


class StylesMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    """
    Dependent on the SettingsMixin being used in the same class
    """
    def set_stylesheet(self):
        """
        Sets the stylesheet for the application based on the current theme
        """
        theme_name = AIRUNNER_DARK_THEME_NAME if self.application_settings.dark_mode_enabled else AIRUNNER_LIGHT_THEME_NAME
        base_dir = Path(os.path.dirname(os.path.realpath(__file__)))
        stylesheet_path = base_dir / theme_name / "styles.qss"

        self.setStyleSheet(
            stylesheet_path.read_text()
            if self.application_settings.override_system_theme
            else ""
        )



from PySide6.QtCore import QObject, Signal, Slot, QThread


class DirectoryWatcher(QObject):
    scan_completed = Signal(bool)

    def __init__(self, base_path: str, scan_function: callable, on_scan_completed: callable):
        super().__init__()
        self.base_path = base_path
        self._scan_function = scan_function
        self._on_scan_completed = on_scan_completed
        self.scan_completed.connect(self.on_scan_completed)

    def run(self):
        while True:
            force_reload = self._scan_function(self.base_path)
            self.scan_completed.emit(force_reload)
            QThread.sleep(1)
    
    @Slot(bool)
    def on_scan_completed(self, force_reload: bool):
        self._on_scan_completed(force_reload)



TINY_AUTOENCODER_FILES_SD = {
    "madebyollin/taesd": [
        "config.json",
        "diffusion_pytorch_model.safetensors",
        "taesd_decoder.safetensors",
        "taesd_encoder.safetensors",
    ]
}

TINY_AUTOENCODER_FILES_SDXL = {
    "madebyollin/sdxl-vae-fp16-fix": [
        "config.json",
        "diffusion_pytorch_model.safetensors",
        "sdxl.vae.safetensors"
    ]
}



import threading
from typing import Dict, Optional

from airunner.enums import SignalCode
from airunner.workers.worker import Worker
from airunner.handlers.llm.llm_model_manager import LLMModelManager
from airunner.settings import AIRUNNER_LLM_ON


class LLMGenerateWorker(Worker):
    def __init__(self, local_agent_class=None):
        self.llm = None
        self.local_agent_class = local_agent_class
        self.signal_handlers = {
            SignalCode.LLM_UNLOAD_SIGNAL: self.on_llm_on_unload_signal,
            SignalCode.LLM_LOAD_SIGNAL: self.on_llm_load_model_signal,
            SignalCode.LLM_CLEAR_HISTORY_SIGNAL: self.on_llm_clear_history_signal,
            SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL: self.on_llm_request_signal,
            SignalCode.RAG_RELOAD_INDEX_SIGNAL: self.on_llm_reload_rag_index_signal,
            SignalCode.ADD_CHATBOT_MESSAGE_SIGNAL: self.on_llm_add_chatbot_response_to_history,
            SignalCode.LOAD_CONVERSATION: self.on_llm_load_conversation,
            SignalCode.INTERRUPT_PROCESS_SIGNAL: self.llm_on_interrupt_process_signal,
            SignalCode.QUIT_APPLICATION: self.on_quit_application_signal,
            SignalCode.CONVERSATION_DELETED: self.on_conversation_deleted_signal,
            SignalCode.SECTION_CHANGED: self.on_section_changed_signal,
            SignalCode.WEB_BROWSER_PAGE_HTML: self.on_web_browser_page_html_signal,
        }
        super().__init__()
        self._llm_thread = None

    def on_conversation_deleted_signal(self, data):
        self.llm.on_conversation_deleted(data)
    
    def on_section_changed_signal(self):
        self.llm.on_section_changed()
    
    def on_web_browser_page_html_signal(self, data):
        if self.llm:
            self.llm.on_web_browser_page_html(data.get("content", ""))

    def on_quit_application_signal(self):
        self.logger.debug("Quitting LLM")
        self.running = False
        if self.llm:
            self.llm.unload()
        if self._llm_thread is not None:
            self._llm_thread.join()

    def on_llm_on_unload_signal(self, data=None):
        if not self.llm:
            return
        data = data or {}
        self.llm.unload()
        callback = data.get("callback", None)
        if callback:
            callback(data)

    def on_llm_load_model_signal(self, data):
        self._load_llm_thread(data)

    def on_llm_clear_history_signal(self, data: Optional[Dict] = None):
        if self.llm:
            self.llm.clear_history(data)

    def on_llm_request_signal(self, message: dict):
        self.add_to_queue(message)

    def llm_on_interrupt_process_signal(self):
        if self.llm:
            self.llm.do_interrupt()

    def on_llm_reload_rag_index_signal(self):
        if self.llm:
            self.llm.reload_rag_engine()

    def on_llm_add_chatbot_response_to_history(self, message):
        self.llm.add_chatbot_response_to_history(message)

    def on_llm_load_conversation(self, message):
        try:
            self.llm.load_conversation(message)
        except Exception as e:
            self.logger.error(f"Error in on_load_conversation: {e}")

    def start_worker_thread(self):
        if self.application_settings.llm_enabled or AIRUNNER_LLM_ON:
            self._load_llm_thread()

    def handle_message(self, message):
        if self.llm:
            self.llm.handle_request(message)

    def _load_llm_thread(self, data=None):
        self._llm_thread = threading.Thread(target=self._load_llm, args=(data,))
        self._llm_thread.start()

    def load(self):
        self._load_llm()

    def _load_llm(self, data=None):
        data = data or {}
        if self.llm is None:
            self.llm = LLMModelManager(local_agent_class=self.local_agent_class)

        self.llm.load()

        callback = data.get("callback", None)
        if callback:
            callback(data)
